import csv
import datetime
import os
import uuid
from unittest import TestCase

import pandas as pd
import testing.postgresql
from mock import Mock
from sqlalchemy import create_engine
from contextlib import contextmanager

from triage.component import metta
from triage.component.architect.feature_group_creator import FeatureGroup
from triage.component.architect.builders import MatrixBuilder
from triage.component.catwalk.db import ensure_db
from triage.component.catwalk.storage import ProjectStorage, HDFMatrixStore
from triage.component.results_schema.schema import Matrix

from .utils import (
    create_schemas,
    create_entity_date_df,
    convert_string_column_to_date,
    TemporaryDirectory,
)

from tests.utils import matrix_metadata_creator


# make some fake features data

states = [
    [0, '2016-01-01', False, True],
    [0, '2016-02-01', False, True],
    [0, '2016-03-01', False, True],
    [0, '2016-04-01', False, True],
    [0, '2016-05-01', False, True],
    [0, '2016-06-01', True, True],
    [1, '2016-01-01', True, False],
    [1, '2016-02-01', True, False],
    [1, '2016-03-01', True, False],
    [1, '2016-04-01', True, False],
    [1, '2016-05-01', True, False],
    [2, '2016-01-01', True, False],
    [2, '2016-02-01', True, True],
    [2, '2016-03-01', True, False],
    [2, '2016-04-01', True, True],
    [2, '2016-05-01', True, False],
    [3, '2016-01-01', False, True],
    [3, '2016-02-01', True, True],
    [3, '2016-03-01', False, True],
    [3, '2016-04-01', True, True],
    [3, '2016-05-01', False, True],
    [4, '2016-01-01', True, True],
    [4, '2016-02-01', True, True],
    [4, '2016-03-01', True, True],
    [4, '2016-04-01', True, True],
    [4, '2016-05-01', True, True],
    [5, '2016-01-01', False, False],
    [5, '2016-02-01', False, False],
    [5, '2016-03-01', False, False],
    [5, '2016-04-01', False, False],
    [5, '2016-05-01', False, False]
]

features0_pre = [
    [0, '2016-01-01', 2, 1],
    [1, '2016-01-01', 1, 2],
    [0, '2016-02-01', 2, 3],
    [1, '2016-02-01', 2, 4],
    [0, '2016-03-01', 3, 3],
    [1, '2016-03-01', 3, 4],
    [0, '2016-04-01', 4, 3],
    [1, '2016-05-01', 5, 4]
]

features1_pre = [
    [2, '2016-01-01', 1, 1],
    [3, '2016-01-01', 1, 2],
    [2, '2016-02-01', 2, 3],
    [3, '2016-02-01', 2, 2],
    [0, '2016-03-01', 3, 3],
    [1, '2016-03-01', 3, 4],
    [2, '2016-03-01', 3, 3],
    [3, '2016-03-01', 3, 4],
    [3, '2016-03-01', 3, 4],
    [0, '2016-03-01', 3, 3],
    [4, '2016-03-01', 1, 4],
    [5, '2016-03-01', 2, 4]
]

# collate will ensure every entity/date combination in the state
# table have an imputed value in the features table, so ensure
# this is true for our test (filling with 9's):
f0_dict = {(r[0], r[1]): r for r in features0_pre}
f1_dict = {(r[0], r[1]): r for r in features1_pre}

for rec in states:
    ent_dt = (rec[0], rec[1])
    f0_dict[ent_dt] = f0_dict.get(ent_dt, list(ent_dt + (9, 9)))
    f1_dict[ent_dt] = f1_dict.get(ent_dt, list(ent_dt + (9, 9)))

features0 = sorted(f0_dict.values(), key=lambda x: (x[1], x[0]))
features1 = sorted(f1_dict.values(), key=lambda x: (x[1], x[0]))

features_tables = [features0, features1]

# make some fake labels data

labels = [
    [0, '2016-02-01', '1 month', 'booking', 'binary', 0],
    [0, '2016-03-01', '1 month', 'booking', 'binary', 0],
    [0, '2016-04-01', '1 month', 'booking', 'binary', 0],
    [0, '2016-05-01', '1 month', 'booking', 'binary', 1],
    [0, '2016-01-01', '1 month', 'ems',     'binary', 0],
    [0, '2016-02-01', '1 month', 'ems',     'binary', 0],
    [0, '2016-03-01', '1 month', 'ems',     'binary', 0],
    [0, '2016-04-01', '1 month', 'ems',     'binary', 0],
    [0, '2016-05-01', '1 month', 'ems',     'binary', 0],
    [1, '2016-01-01', '1 month', 'booking', 'binary', 0],
    [1, '2016-02-01', '1 month', 'booking', 'binary', 0],
    [1, '2016-03-01', '1 month', 'booking', 'binary', 0],
    [1, '2016-04-01', '1 month', 'booking', 'binary', 0],
    [1, '2016-05-01', '1 month', 'booking', 'binary', 1],
    [1, '2016-01-01', '1 month', 'ems',     'binary', 0],
    [1, '2016-02-01', '1 month', 'ems',     'binary', 0],
    [1, '2016-03-01', '1 month', 'ems',     'binary', 0],
    [1, '2016-04-01', '1 month', 'ems',     'binary', 0],
    [1, '2016-05-01', '1 month', 'ems',     'binary', 0],
    [2, '2016-01-01', '1 month', 'booking', 'binary', 0],
    [2, '2016-02-01', '1 month', 'booking', 'binary', 0],
    [2, '2016-03-01', '1 month', 'booking', 'binary', 1],
    [2, '2016-04-01', '1 month', 'booking', 'binary', 0],
    [2, '2016-05-01', '1 month', 'booking', 'binary', 1],
    [2, '2016-01-01', '1 month', 'ems',     'binary', 0],
    [2, '2016-02-01', '1 month', 'ems',     'binary', 0],
    [2, '2016-03-01', '1 month', 'ems',     'binary', 0],
    [2, '2016-04-01', '1 month', 'ems',     'binary', 0],
    [2, '2016-05-01', '1 month', 'ems',     'binary', 1],
    [3, '2016-01-01', '1 month', 'booking', 'binary', 0],
    [3, '2016-02-01', '1 month', 'booking', 'binary', 0],
    [3, '2016-03-01', '1 month', 'booking', 'binary', 1],
    [3, '2016-04-01', '1 month', 'booking', 'binary', 0],
    [3, '2016-05-01', '1 month', 'booking', 'binary', 1],
    [3, '2016-01-01', '1 month', 'ems',     'binary', 0],
    [3, '2016-02-01', '1 month', 'ems',     'binary', 0],
    [3, '2016-03-01', '1 month', 'ems',     'binary', 0],
    [3, '2016-04-01', '1 month', 'ems',     'binary', 1],
    [3, '2016-05-01', '1 month', 'ems',     'binary', 0],
    [4, '2016-01-01', '1 month', 'booking', 'binary', 1],
    [4, '2016-02-01', '1 month', 'booking', 'binary', 0],
    [4, '2016-03-01', '1 month', 'booking', 'binary', 0],
    [4, '2016-04-01', '1 month', 'booking', 'binary', 0],
    [4, '2016-05-01', '1 month', 'booking', 'binary', 0],
    [4, '2016-01-01', '1 month', 'ems',     'binary', 0],
    [4, '2016-02-01', '1 month', 'ems',     'binary', 1],
    [4, '2016-03-01', '1 month', 'ems',     'binary', 0],
    [4, '2016-04-01', '1 month', 'ems',     'binary', 1],
    [4, '2016-05-01', '1 month', 'ems',     'binary', 1],
    [5, '2016-01-01', '1 month', 'booking', 'binary', 1],
    [5, '2016-02-01', '1 month', 'booking', 'binary', 0],
    [5, '2016-03-01', '1 month', 'booking', 'binary', 0],
    [5, '2016-04-01', '1 month', 'booking', 'binary', 0],
    [5, '2016-05-01', '1 month', 'booking', 'binary', 0],
    [5, '2016-01-01', '1 month', 'ems',     'binary', 0],
    [5, '2016-02-01', '1 month', 'ems',     'binary', 1],
    [5, '2016-03-01', '1 month', 'ems',     'binary', 0],
    [5, '2016-04-01', '1 month', 'ems',     'binary', 0],
    [5, '2016-05-01', '1 month', 'ems',     'binary', 0],
    [0, '2016-02-01', '3 month', 'booking', 'binary', 0],
    [0, '2016-03-01', '3 month', 'booking', 'binary', 0],
    [0, '2016-04-01', '3 month', 'booking', 'binary', 0],
    [0, '2016-05-01', '3 month', 'booking', 'binary', 1],
    [0, '2016-01-01', '3 month', 'ems',     'binary', 0],
    [0, '2016-02-01', '3 month', 'ems',     'binary', 0],
    [0, '2016-03-01', '3 month', 'ems',     'binary', 0],
    [0, '2016-04-01', '3 month', 'ems',     'binary', 0],
    [0, '2016-05-01', '3 month', 'ems',     'binary', 0],
    [1, '2016-01-01', '3 month', 'booking', 'binary', 0],
    [1, '2016-02-01', '3 month', 'booking', 'binary', 0],
    [1, '2016-03-01', '3 month', 'booking', 'binary', 0],
    [1, '2016-04-01', '3 month', 'booking', 'binary', 0],
    [1, '2016-05-01', '3 month', 'booking', 'binary', 1],
    [1, '2016-01-01', '3 month', 'ems',     'binary', 0],
    [1, '2016-02-01', '3 month', 'ems',     'binary', 0],
    [1, '2016-03-01', '3 month', 'ems',     'binary', 0],
    [1, '2016-04-01', '3 month', 'ems',     'binary', 0],
    [1, '2016-05-01', '3 month', 'ems',     'binary', 0],
    [2, '2016-01-01', '3 month', 'booking', 'binary', 0],
    [2, '2016-02-01', '3 month', 'booking', 'binary', 0],
    [2, '2016-03-01', '3 month', 'booking', 'binary', 1],
    [2, '2016-04-01', '3 month', 'booking', 'binary', 0],
    [2, '2016-05-01', '3 month', 'booking', 'binary', 1],
    [2, '2016-01-01', '3 month', 'ems',     'binary', 0],
    [2, '2016-02-01', '3 month', 'ems',     'binary', 0],
    [2, '2016-03-01', '3 month', 'ems',     'binary', 0],
    [2, '2016-04-01', '3 month', 'ems',     'binary', 0],
    [2, '2016-05-01', '3 month', 'ems',     'binary', 1],
    [3, '2016-01-01', '3 month', 'booking', 'binary', 0],
    [3, '2016-02-01', '3 month', 'booking', 'binary', 0],
    [3, '2016-03-01', '3 month', 'booking', 'binary', 1],
    [3, '2016-04-01', '3 month', 'booking', 'binary', 0],
    [3, '2016-05-01', '3 month', 'booking', 'binary', 1],
    [3, '2016-01-01', '3 month', 'ems',     'binary', 0],
    [3, '2016-02-01', '3 month', 'ems',     'binary', 0],
    [3, '2016-03-01', '3 month', 'ems',     'binary', 0],
    [3, '2016-04-01', '3 month', 'ems',     'binary', 1],
    [3, '2016-05-01', '3 month', 'ems',     'binary', 0],
    [3, '2016-05-01', '3 month', 'ems',     'binary', 0],
    [4, '2016-01-01', '3 month', 'booking', 'binary', 0],
    [4, '2016-02-01', '3 month', 'booking', 'binary', 0],
    [4, '2016-03-01', '3 month', 'booking', 'binary', 1],
    [4, '2016-04-01', '3 month', 'booking', 'binary', 0],
    [4, '2016-05-01', '3 month', 'booking', 'binary', 1],
    [4, '2016-01-01', '3 month', 'ems',     'binary', 0],
    [4, '2016-02-01', '3 month', 'ems',     'binary', 0],
    [4, '2016-03-01', '3 month', 'ems',     'binary', 0],
    [4, '2016-04-01', '3 month', 'ems',     'binary', 0],
    [4, '2016-05-01', '3 month', 'ems',     'binary', 1],
    [5, '2016-01-01', '3 month', 'booking', 'binary', 0],
    [5, '2016-02-01', '3 month', 'booking', 'binary', 0],
    [5, '2016-03-01', '3 month', 'booking', 'binary', 1],
    [5, '2016-04-01', '3 month', 'booking', 'binary', 0],
    [5, '2016-05-01', '3 month', 'booking', 'binary', 1],
    [5, '2016-01-01', '3 month', 'ems',     'binary', 0],
    [5, '2016-02-01', '3 month', 'ems',     'binary', 0],
    [5, '2016-03-01', '3 month', 'ems',     'binary', 0],
    [5, '2016-04-01', '3 month', 'ems',     'binary', 1],
    [5, '2016-05-01', '3 month', 'ems',     'binary', 0]
]

label_name = 'booking'
label_type = 'binary'

db_config = {
    'features_schema_name': 'features',
    'labels_schema_name': 'labels',
    'labels_table_name': 'labels',
    'sparse_state_table_name': 'staging.sparse_states',
}


@contextmanager
def get_matrix_storage_engine():
    with TemporaryDirectory() as temp_dir:
        yield ProjectStorage(temp_dir).matrix_storage_engine()


def test_query_to_df():
    """ Test the write_to_csv function by checking whether the csv contains the
    correct number of lines.
    """
    with testing.postgresql.Postgresql() as postgresql:
        # create an engine and generate a table with fake feature data
        engine = create_engine(postgresql.url())
        create_schemas(
            engine=engine,
            features_tables=features_tables,
            labels=labels,
            states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                engine=engine,
            )

            # for each table, check that corresponding csv has the correct # of rows
            for table in features_tables:
                df = builder.query_to_df(
                    '''
                        select *
                        from features.features{}
                    '''.format(features_tables.index(table))
                )
                assert(len(df) == len(table))


def test_make_entity_date_table():
    """ Test that the make_entity_date_table function contains the correct
    values.
    """
    dates = [datetime.datetime(2016, 1, 1, 0, 0),
             datetime.datetime(2016, 2, 1, 0, 0),
             datetime.datetime(2016, 3, 1, 0, 0)]

    # make a dataframe of entity ids and dates to test against
    ids_dates = create_entity_date_df(
        labels=labels,
        states=states,
        as_of_dates=dates,
        state_one=True,
        state_two=True,
        label_name='booking',
        label_type='binary',
        label_timespan='1 month'
    )

    with testing.postgresql.Postgresql() as postgresql:
        # create an engine and generate a table with fake feature data
        engine = create_engine(postgresql.url())
        create_schemas(
            engine=engine,
            features_tables=features_tables,
            labels=labels,
            states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                engine=engine
            )
            engine.execute(
                'CREATE TABLE features.tmp_entity_date (a int, b date);'
            )
            # call the function to test the creation of the table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type='binary',
                label_name='booking',
                state='state_one AND state_two',
                matrix_uuid='my_uuid',
                matrix_type='train',
                label_timespan='1 month'
            )

            # read in the table
            result = pd.read_sql(
                "select * from features.{} order by entity_id, as_of_date"
                .format(entity_date_table_name),
                engine
            )
            # compare the table to the test dataframe
            test = (result == ids_dates)
            assert(test.all().all())


def test_make_entity_date_table_include_missing_labels():
    """ Test that the make_entity_date_table function contains the correct
    values.
    """
    dates = [datetime.datetime(2016, 1, 1, 0, 0),
             datetime.datetime(2016, 2, 1, 0, 0),
             datetime.datetime(2016, 3, 1, 0, 0),
             datetime.datetime(2016, 6, 1, 0, 0)]

    # same as the other make_entity_date_label test except there is an extra date, 2016-06-01
    # entity 0 is included in this date via the states table, but has no label

    # make a dataframe of entity ids and dates to test against
    ids_dates = create_entity_date_df(
        labels=labels,
        states=states,
        as_of_dates=dates,
        state_one=True,
        state_two=True,
        label_name='booking',
        label_type='binary',
        label_timespan='1 month'
    )
    # this line adds the new entity-date combo as an expected one
    ids_dates = ids_dates.append({'entity_id': 0, 'as_of_date': datetime.date(2016, 6, 1)}, ignore_index=True)

    with testing.postgresql.Postgresql() as postgresql:
        # create an engine and generate a table with fake feature data
        engine = create_engine(postgresql.url())
        create_schemas(
            engine=engine,
            features_tables=features_tables,
            labels=labels,
            states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                include_missing_labels_in_train_as=False,
                engine=engine
            )
            engine.execute(
                'CREATE TABLE features.tmp_entity_date (a int, b date);'
            )
            # call the function to test the creation of the table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type='binary',
                label_name='booking',
                state='state_one AND state_two',
                matrix_uuid='my_uuid',
                matrix_type='train',
                label_timespan='1 month'
            )

            # read in the table
            result = pd.read_sql(
                "select * from features.{} order by entity_id, as_of_date"
                .format(entity_date_table_name),
                engine
            )

            # compare the table to the test dataframe
            assert sorted(result.values.tolist()) == sorted(ids_dates.values.tolist())


def test_load_features_data():
    dates = [datetime.datetime(2016, 1, 1, 0, 0),
             datetime.datetime(2016, 2, 1, 0, 0)]

    # make dataframe for entity ids and dates
    ids_dates = create_entity_date_df(
        labels=labels,
        states=states,
        as_of_dates=dates,
        state_one=True,
        state_two=True,
        label_name='booking',
        label_type='binary',
        label_timespan='1 month'
    )

    features = [['f1', 'f2'], ['f3', 'f4']]
    # make dataframes of features to test against
    features_dfs = []
    for i, table in enumerate(features_tables):
        cols = ['entity_id', 'as_of_date'] + features[i]
        temp_df = pd.DataFrame(table, columns=cols)
        temp_df['as_of_date'] = convert_string_column_to_date(temp_df['as_of_date'])
        features_dfs.append(
            ids_dates.merge(
                right=temp_df,
                how='left',
                on=['entity_id', 'as_of_date']
            ).set_index(['entity_id', 'as_of_date'])
        )

    # create an engine and generate a table with fake feature data
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_schemas(
            engine=engine,
            features_tables=features_tables,
            labels=labels,
            states=states
        )

        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                engine=engine,
            )

            # make the entity-date table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type='binary',
                label_name='booking',
                state='state_one AND state_two',
                matrix_type='train',
                matrix_uuid='my_uuid',
                label_timespan='1 month'
            )

            feature_dictionary = dict(
                ('features{}'.format(i), feature_list) for i, feature_list in enumerate(features)
            )

            returned_features_dfs = builder.load_features_data(
                as_of_times=dates,
                feature_dictionary=feature_dictionary,
                entity_date_table_name=entity_date_table_name,
                matrix_uuid='my_uuid'
            )

            # get the queries and test them
            for result, df in zip(returned_features_dfs, features_dfs):
                test = (result == df)
                assert(test.all().all())


def test_load_labels_data():
    """ Test the load_labels_data function by checking whether the query
    produces the correct labels
    """
    # set up labeling config variables
    dates = [datetime.datetime(2016, 1, 1, 0, 0),
             datetime.datetime(2016, 2, 1, 0, 0)]

    # make a dataframe of labels to test against
    labels_df = pd.DataFrame(
        labels,
        columns=[
            'entity_id',
            'as_of_date',
            'label_timespan',
            'label_name',
            'label_type',
            'label'
        ]
    )

    labels_df['as_of_date'] = convert_string_column_to_date(labels_df['as_of_date'])
    labels_df.set_index(['entity_id', 'as_of_date'])

    # create an engine and generate a table with fake feature data
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_schemas(
            engine,
            features_tables,
            labels,
            states
        )
        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                engine=engine,
            )

            # make the entity-date table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type='binary',
                label_name='booking',
                state='state_one AND state_two',
                matrix_type='train',
                matrix_uuid='my_uuid',
                label_timespan='1 month'
            )

            result = builder.load_labels_data(
                label_name=label_name,
                label_type=label_type,
                label_timespan='1 month',
                matrix_uuid='my_uuid',
                entity_date_table_name=entity_date_table_name,
            )
            df = pd.DataFrame.from_dict({
                'entity_id': [2, 3, 4, 4],
                'as_of_date': [dates[1], dates[1], dates[0], dates[1]],
                'booking': [0, 0, 1, 0],
            }).set_index(['entity_id', 'as_of_date'])

            test = (result == df)
            assert(test.all().all())


def test_load_labels_data_include_missing_labels_as_false():
    """ Test the load_labels_data function by checking whether the query
    produces the correct labels
    """
    # set up labeling config variables
    dates = [datetime.datetime(2016, 1, 1, 0, 0),
             datetime.datetime(2016, 2, 1, 0, 0),
             datetime.datetime(2016, 6, 1, 0, 0)]

    # same as the other load_labels_data test, except we include an extra date, 2016-06-01
    # this date does have entity 0 included via the states table, but no labels

    # make a dataframe of labels to test against
    labels_df = pd.DataFrame(
        labels,
        columns=[
            'entity_id',
            'as_of_date',
            'label_timespan',
            'label_name',
            'label_type',
            'label'
        ]
    )

    labels_df['as_of_date'] = convert_string_column_to_date(labels_df['as_of_date'])
    labels_df.set_index(['entity_id', 'as_of_date'])

    # create an engine and generate a table with fake feature data
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        create_schemas(
            engine,
            features_tables,
            labels,
            states
        )
        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                engine=engine,
                include_missing_labels_in_train_as=False,
            )

            # make the entity-date table
            entity_date_table_name = builder.make_entity_date_table(
                as_of_times=dates,
                label_type='binary',
                label_name='booking',
                state='state_one AND state_two',
                matrix_type='train',
                matrix_uuid='my_uuid',
                label_timespan='1 month'
            )

            result = builder.load_labels_data(
                label_name=label_name,
                label_type=label_type,
                label_timespan='1 month',
                matrix_uuid='my_uuid',
                entity_date_table_name=entity_date_table_name,
            )
            df = pd.DataFrame.from_dict({
                'entity_id': [0, 2, 3, 4, 4],
                'as_of_date': [dates[2], dates[1], dates[1], dates[0], dates[1]],
                'booking': [0, 0, 0, 1, 0],
            }).set_index(['entity_id', 'as_of_date'])
            # the first row would not be here if we had not configured the Builder
            # to include missing labels as false

            test = (result == df)
            assert(test.all().all())

class TestMergeFeatureCSVs(TestCase):
    def test_badinput(self):
        """We assert column names, so replacing 'date' with 'as_of_date'
        should result in an error"""
        with get_matrix_storage_engine() as matrix_storage_engine:
            builder = MatrixBuilder(
                db_config=db_config,
                matrix_storage_engine=matrix_storage_engine,
                engine=None,
            )
            dataframes = [
                pd.DataFrame.from_records([
                    (1, 3, 3),
                    (4, 5, 6),
                    (7, 8, 9),
                ], columns=('entity_id', 'date', 'f1'), index=['entity_id', 'date']),
                pd.DataFrame.from_records([
                    (1, 2, 3),
                    (4, 5, 9),
                    (7, 8, 15),
                ], columns=('entity_id', 'date', 'f3'), index=['entity_id', 'date']),
                pd.DataFrame.from_records([
                    (1, 2, 2),
                    (4, 5, 20),
                    (7, 8, 56),
                ], columns=('entity_id', 'date', 'f3'), index=['entity_id', 'date']),
            ]

            with self.assertRaises(ValueError):
                builder.merge_feature_csvs(dataframes, matrix_uuid='1234')


class TestBuildMatrix(TestCase):
    @property
    def good_metadata(self):
        return {
            'matrix_id': 'hi',
            'state': 'state_one AND state_two',
            'label_name': 'booking',
            'end_time': datetime.datetime(2016, 3, 1, 0, 0),
            'feature_start_time': datetime.datetime(2016, 1, 1, 0, 0),
            'label_timespan': '1 month',
            'max_training_history': '1 month',
            'test_duration': '1 month',
            'indices': ['entity_id', 'as_of_date'],
        }

    @property
    def good_feature_dictionary(self):
        return FeatureGroup(
            name='mygroup',
            features_by_table={
                'features0': ['f1', 'f2'],
                'features1': ['f3', 'f4'],
            }
        )

    @property
    def good_dates(self):
        return [
            datetime.datetime(2016, 1, 1, 0, 0),
            datetime.datetime(2016, 2, 1, 0, 0),
            datetime.datetime(2016, 3, 1, 0, 0)
        ]

    def test_train_matrix(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states
            )


            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    engine=engine
                )
                uuid = metta.generate_uuid(self.good_metadata)
                builder.build_matrix(
                    as_of_times=self.good_dates,
                    label_name='booking',
                    label_type='binary',
                    feature_dictionary=self.good_feature_dictionary,
                    matrix_metadata=self.good_metadata,
                    matrix_uuid=uuid,
                    matrix_type='train'
                )
                assert len(matrix_storage_engine.get_store(uuid).matrix) == 5

    def test_test_matrix(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    engine=engine
                )

                uuid = metta.generate_uuid(self.good_metadata)
                builder.build_matrix(
                    as_of_times=self.good_dates,
                    label_name='booking',
                    label_type='binary',
                    feature_dictionary=self.good_feature_dictionary,
                    matrix_metadata=self.good_metadata,
                    matrix_uuid=uuid,
                    matrix_type='test'
                )

                assert len(matrix_storage_engine.get_store(uuid).matrix) == 5

    def test_hdf_matrix(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                matrix_storage_engine.matrix_storage_class = HDFMatrixStore
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    engine=engine
                )

                uuid = metta.generate_uuid(self.good_metadata)
                builder.build_matrix(
                    as_of_times=self.good_dates,
                    label_name='booking',
                    label_type='binary',
                    feature_dictionary=self.good_feature_dictionary,
                    matrix_metadata=self.good_metadata,
                    matrix_uuid=uuid,
                    matrix_type='test'
                )

                assert len(matrix_storage_engine.get_store(uuid).matrix) == 5

    def test_nullcheck(self):
        f0_dict = {(r[0], r[1]): r for r in features0_pre}
        f1_dict = {(r[0], r[1]): r for r in features1_pre}

        features0 = sorted(f0_dict.values(), key=lambda x: (x[1], x[0]))
        features1 = sorted(f1_dict.values(), key=lambda x: (x[1], x[0]))

        features_tables = [features0, features1]

        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states
            )

            dates = [datetime.datetime(2016, 1, 1, 0, 0),
                     datetime.datetime(2016, 2, 1, 0, 0),
                     datetime.datetime(2016, 3, 1, 0, 0)]

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    engine=engine
                )

                feature_dictionary = {
                    'features0': ['f1', 'f2'],
                    'features1': ['f3', 'f4'],
                }
                matrix_metadata = {
                    'matrix_id': 'hi',
                    'state': 'state_one AND state_two',
                    'label_name': 'booking',
                    'end_time': datetime.datetime(2016, 3, 1, 0, 0),
                    'feature_start_time': datetime.datetime(2016, 1, 1, 0, 0),
                    'label_timespan': '1 month',
                    'test_duration': '1 month',
                    'indices': ['entity_id', 'as_of_date'],
                }
                uuid = metta.generate_uuid(matrix_metadata)
                with self.assertRaises(ValueError):
                    builder.build_matrix(
                        as_of_times=dates,
                        label_name='booking',
                        label_type='binary',
                        feature_dictionary=feature_dictionary,
                        matrix_metadata=matrix_metadata,
                        matrix_uuid=uuid,
                        matrix_type='test'
                    )

    def test_replace_false_rerun(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states
            )

            dates = [datetime.datetime(2016, 1, 1, 0, 0),
                     datetime.datetime(2016, 2, 1, 0, 0),
                     datetime.datetime(2016, 3, 1, 0, 0)]

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    engine=engine,
                    replace=False
                )

                feature_dictionary = {
                    'features0': ['f1', 'f2'],
                    'features1': ['f3', 'f4'],
                }
                matrix_metadata = {
                    'matrix_id': 'hi',
                    'state': 'state_one AND state_two',
                    'label_name': 'booking',
                    'end_time': datetime.datetime(2016, 3, 1, 0, 0),
                    'feature_start_time': datetime.datetime(2016, 1, 1, 0, 0),
                    'label_timespan': '1 month',
                    'test_duration': '1 month',
                    'indices': ['entity_id', 'as_of_date'],
                }
                uuid = metta.generate_uuid(matrix_metadata)
                builder.build_matrix(
                    as_of_times=dates,
                    label_name='booking',
                    label_type='binary',
                    feature_dictionary=feature_dictionary,
                    matrix_metadata=matrix_metadata,
                    matrix_uuid=uuid,
                    matrix_type='test'
                )

                assert len(matrix_storage_engine.get_store(uuid).matrix) == 5
                # rerun
                builder.make_entity_date_table = Mock()
                builder.build_matrix(
                    as_of_times=dates,
                    label_name='booking',
                    label_type='binary',
                    feature_dictionary=feature_dictionary,
                    matrix_metadata=matrix_metadata,
                    matrix_uuid=uuid,
                    matrix_type='test'
                )
                assert not builder.make_entity_date_table.called

    def test_replace_true_rerun(self):
        with testing.postgresql.Postgresql() as postgresql:
            # create an engine and generate a table with fake feature data
            engine = create_engine(postgresql.url())
            ensure_db(engine)
            create_schemas(
                engine=engine,
                features_tables=features_tables,
                labels=labels,
                states=states
            )
            matrix_metadata = matrix_metadata_creator(
                state='state_one and state_two',
                test_duration='1month'
            )

            dates = [datetime.datetime(2016, 1, 1, 0, 0),
                     datetime.datetime(2016, 2, 1, 0, 0),
                     datetime.datetime(2016, 3, 1, 0, 0)]

            feature_dictionary = {
                'features0': ['f1', 'f2'],
                'features1': ['f3', 'f4'],
            }
            uuid = metta.generate_uuid(matrix_metadata)
            build_args = dict(
                as_of_times=dates,
                label_name='booking',
                label_type='binary',
                feature_dictionary=feature_dictionary,
                matrix_metadata=matrix_metadata,
                matrix_uuid=uuid,
                matrix_type='test'
            )

            with get_matrix_storage_engine() as matrix_storage_engine:
                builder = MatrixBuilder(
                    db_config=db_config,
                    matrix_storage_engine=matrix_storage_engine,
                    engine=engine,
                    replace=True
                )

                builder.build_matrix(**build_args)

                assert len(matrix_storage_engine.get_store(uuid).matrix) == 5
                assert builder.sessionmaker().query(Matrix).get(uuid)
                # rerun
                builder.build_matrix(**build_args)
                assert len(matrix_storage_engine.get_store(uuid).matrix) == 5
                assert builder.sessionmaker().query(Matrix).get(uuid)



import tempfile
import random
from contextlib import contextmanager
import os

import pandas as pd
import requests
from tqdm import tqdm

@contextmanager
def download(url):
    "download `url` to a file, returning the file name"
    with tempfile.NamedTemporaryFile(mode="wb") as f:
        response = requests.get(url, stream=True)
        for data in tqdm(response.iter_content()):
            f.write(data)
        f.flush()
        yield f.name

def create_subset(src, dest, n=250):
    "Given a csv file `src`, create a subset `dest` with `n` unique entities"
    df = pd.read_csv(src)
    lics = pd.unique(df["License #"])
    sublics = lics[random.sample(range(0,len(lics)), n)]
    subset = df[df["License #"].isin(sublics)]
    # Make the column names a little more readable
    subset.columns = map(clean_column_name, subset.columns)
    subset.to_csv(dest, index=False)

def clean_column_name(col):
    col = col.lower()
    col = col.replace(' ', '_')
    col = col.replace('#', 'no')
    return re.sub('[\W]+', '', col)

if __name__ == '__main__':
    # download the entire Chicago restaurant inspections CSV file
    with download("https://data.cityofchicago.org/api/views/4ijn-s7e5/rows.csv?accessType=DOWNLOAD") as f:
        create_subset(f, "food_inspections_subset.csv")



"""empty message

Revision ID: 7d57d1cf3429
Revises: 72ac5cbdca05
Create Date: 2017-11-06 11:34:23.046005

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '7d57d1cf3429'
down_revision = '72ac5cbdca05'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('evaluations', 'example_frequency', new_column_name='as_of_date_frequency')
    op.alter_column('models', 'train_label_window', new_column_name='training_label_timespan')
    op.alter_column('predictions', 'test_label_window', new_column_name='test_label_timespan')
    op.alter_column('list_predictions', 'test_label_window', new_column_name='test_label_timespan')


def downgrade():
    op.alter_column('evaluations', 'as_of_date_frequency', new_column_name='example_frequency')
    op.alter_column('models', 'training_label_timespan', new_column_name='train_label_window')
    op.alter_column('predictions', 'test_label_timespan', new_column_name='test_label_window')
    op.alter_column('list_predictions', 'test_label_timespan', new_column_name='test_label_window')



# Avoid circular import (required by base)
CONFIG_VERSION = 'v5'  # noqa: E402

from .base import ExperimentBase
from .multicore import MultiCoreExperiment
from .singlethreaded import SingleThreadedExperiment

__all__ = (
    'ExperimentBase',
    'MultiCoreExperiment',
    'SingleThreadedExperiment',
)



import functools
import logging
import time

import numpy
from sqlalchemy.orm import sessionmaker

from . import metrics
from .utils import db_retry, sort_predictions_and_labels


def generate_binary_at_x(test_predictions, x_value, unit='top_n'):
    """Generate subset of predictions based on top% or absolute

    Args:
        test_predictions (list) A list of predictions, sorted by risk desc
        x_value (int) The percentile or absolute value desired
        unit (string, default 'top_n') The subsetting method desired,
            either percentile or top_n

    Returns: (list) The predictions subset
    """
    if unit == 'percentile':
        cutoff_index = int(len(test_predictions) * (x_value / 100.00))
    else:
        cutoff_index = x_value
    test_predictions_binary = [
        1 if x < cutoff_index else 0
        for x in range(len(test_predictions))
    ]
    return test_predictions_binary


class ModelEvaluator(object):
    """An object that can score models based on its known metrics"""

    """Available metric calculation functions

    Each value is expected to be a function that takes in the following params
    (predictions_proba, predictions_binary, labels, parameters)
    and return a numeric score
    """
    available_metrics = {
        'precision@': metrics.precision,
        'recall@': metrics.recall,
        'fbeta@': metrics.fbeta,
        'f1': metrics.f1,
        'accuracy': metrics.accuracy,
        'roc_auc': metrics.roc_auc,
        'average precision score': metrics.avg_precision,
        'true positives@': metrics.true_positives,
        'true negatives@': metrics.true_negatives,
        'false positives@': metrics.false_positives,
        'false negatives@': metrics.false_negatives,
        'fpr@': metrics.fpr,
    }

    def __init__(
        self,
        testing_metric_groups,
        training_metric_groups,
        db_engine,
        sort_seed=None,
        custom_metrics=None
    ):
        """
        Args:
            testing_metric_groups (list) A list of groups of metric/configurations
                to use for evaluating all given models

                Each entry is a dict, with a list of metrics, and potentially
                    thresholds and parameter lists. Each metric is expected to
                    be a key in self.available_metrics

                Examples:

                testing_metric_groups = [{
                    'metrics': ['precision@', 'recall@'],
                    'thresholds': {
                        'percentiles': [5.0, 10.0],
                        'top_n': [5, 10]
                    }
                }, {
                    'metrics': ['f1'],
                }, {
                    'metrics': ['fbeta@'],
                    'parameters': [{'beta': 0.75}, {'beta': 1.25}]
                }]
            training_metric_groups (list) metrics to be calculated on training set,
                in the same form as testing_metric_groups
            db_engine (sqlalchemy.engine)
            custom_metrics (dict) Functions to generate metrics
                not available by default
                Each function is expected take in the following params:
                (predictions_proba, predictions_binary, labels, parameters)
                and return a numeric score
        """
        self.testing_metric_groups = testing_metric_groups
        self.training_metric_groups = training_metric_groups
        self.db_engine = db_engine
        self.sort_seed = sort_seed or int(time.time())
        if custom_metrics:
            self._validate_metrics(custom_metrics)
            self.available_metrics.update(custom_metrics)

    @property
    def sessionmaker(self):
        return sessionmaker(bind=self.db_engine)

    def _validate_metrics(
        self,
        custom_metrics
    ):
        for name, met in custom_metrics.items():
            if not hasattr(met, 'greater_is_better'):
                raise ValueError("Custom metric {} missing greater_is_better "
                                 "attribute".format(name))
            elif met.greater_is_better not in (True, False):
                raise ValueError("For custom metric {} greater_is_better must be "
                                 "boolean True or False".format(name))

    def _build_parameter_string(
        self,
        threshold_unit,
        threshold_value,
        parameter_combination,
        threshold_specified_by_user
    ):
        """Encode the metric parameters and threshold into a short, human-parseable string

        Examples are: '100_abs', '5_pct'

        Args:
            threshold_unit (string) the type of threshold, either 'percentile' or 'top_n'
            threshold_value (int) the numeric threshold,
            parameter_combination (dict) The non-threshold parameter keys and values used
                Usually this will be empty, but an example would be {'beta': 0.25}

        Returns: (string) A short, human-parseable string
        """
        full_params = parameter_combination.copy()
        if threshold_specified_by_user:
            short_threshold_unit = 'pct' if threshold_unit == 'percentile' else 'abs'
            full_params[short_threshold_unit] = threshold_value
        parameter_string = '/'.join([
            '{}_{}'.format(val, key)
            for key, val in full_params.items()
        ])
        return parameter_string

    def _filter_nan_labels(self, predicted_classes, labels):
        """Filter missing labels and their corresponding predictions

        Args:
            predicted_classes (list) Predicted binary classes, of same length as labels
            labels (list) Labels, maybe containing NaNs

        Returns: (tuple) Copies of the input lists, with NaN labels removed
        """
        labels = numpy.array(labels)
        predicted_classes = numpy.array(predicted_classes)
        nan_mask = numpy.isfinite(labels)
        return (
            (predicted_classes[nan_mask]).tolist(),
            (labels[nan_mask]).tolist()
        )

    def _evaluations_for_threshold(
        self,
        metrics,
        parameters,
        predictions_proba,
        labels,
        evaluation_table_obj,
        threshold_unit,
        threshold_value,
        threshold_specified_by_user=True,
    ):
        """Generate evaluations for a given threshold in a metric group,
        and create ORM objects to hold them

        Args:
            metrics (list) names of metric to compute
            parameters (list) dicts holding parameters to pass to metrics
            predictions_proba (list) Probability predictions
            labels (list) True labels (may have NaNs)
            threshold_unit (string) the type of threshold, either 'percentile' or 'top_n'
            threshold_value (int) the numeric threshold,
            threshold_specified_by_user (bool) Whether or not there was any threshold
                specified by the user. Defaults to True
            evaluation_table_obj (schema.TestEvaluation or TrainEvaluation)
                specifies to which table to add the evaluations

        Returns: (list) results_schema.TrainEvaluation or TestEvaluation objects
        Raises: UnknownMetricError if a given metric is not present in
            self.available_metrics
        """

        # using threshold configuration, convert probabilities to predicted classes
        predicted_classes = generate_binary_at_x(
            predictions_proba,
            threshold_value,
            unit=threshold_unit
        )
        # filter out null labels
        predicted_classes_with_labels, present_labels = self._filter_nan_labels(
            predicted_classes,
            labels,
        )
        num_labeled_examples = len(present_labels)
        num_labeled_above_threshold = predicted_classes_with_labels.count(1)
        num_positive_labels = present_labels.count(1)
        evaluations = []
        for metric in metrics:
            if metric not in self.available_metrics:
                raise metrics.UnknownMetricError()

            for parameter_combination in parameters:
                value = self.available_metrics[metric](
                    predictions_proba,
                    predicted_classes_with_labels,
                    present_labels,
                    parameter_combination
                )

                # convert the thresholds/parameters into something
                # more readable
                parameter_string = self._build_parameter_string(
                    threshold_unit=threshold_unit,
                    threshold_value=threshold_value,
                    parameter_combination=parameter_combination,
                    threshold_specified_by_user=threshold_specified_by_user
                )

                logging.info(
                    'Evaluations for %s%s, labeled examples %s '
                    'above threshold %s, positive labels %s, value %s',
                    metric,
                    parameter_string,
                    num_labeled_examples,
                    num_labeled_above_threshold,
                    num_positive_labels,
                    value
                )
                evaluations.append(evaluation_table_obj(
                    metric=metric,
                    parameter=parameter_string,
                    value=value,
                    num_labeled_examples=num_labeled_examples,
                    num_labeled_above_threshold=num_labeled_above_threshold,
                    num_positive_labels=num_positive_labels,
                    sort_seed=self.sort_seed
                ))
        return evaluations

    def _evaluations_for_group(
        self,
        group,
        predictions_proba_sorted,
        labels_sorted,
        evaluation_table_obj,
    ):
        """Generate evaluations for a given metric group, and create ORM objects to hold them

        Args:
            group (dict) A configuration dictionary for the group.
                Should contain the key 'metrics', and optionally 'parameters' or 'thresholds'
            predictions_proba (list) Probability predictions
            labels (list) True labels (may have NaNs)

        Returns: (list) results_schema.Evaluation objects
        """
        logging.info('Creating evaluations for metric group %s', group)
        parameters = group.get('parameters', [{}])
        generate_evaluations = functools.partial(
            self._evaluations_for_threshold,
            metrics=group['metrics'],
            parameters=parameters,
            predictions_proba=predictions_proba_sorted,
            labels=labels_sorted,
            evaluation_table_obj=evaluation_table_obj
        )
        evaluations = []
        if 'thresholds' not in group:
            logging.info('Not a thresholded group, generating evaluation based on all predictions')
            evaluations = evaluations + generate_evaluations(
                threshold_unit='percentile',
                threshold_value=100,
                threshold_specified_by_user=False
            )

        for pct_thresh in group.get('thresholds', {}).get('percentiles', []):
            logging.info('Processing percent threshold %s', pct_thresh)
            evaluations = evaluations + generate_evaluations(
                threshold_unit='percentile',
                threshold_value=pct_thresh
            )

        for abs_thresh in group.get('thresholds', {}).get('top_n', []):
            logging.info('Processing absolute threshold %s', abs_thresh)
            evaluations = evaluations + generate_evaluations(
                threshold_unit='top_n',
                threshold_value=abs_thresh
            )
        return evaluations

    def evaluate(self, predictions_proba, matrix_store, model_id):
        """Evaluate a model based on predictions, and save the results

        Args:
            predictions_proba (numpy.array) List of prediction probabilities
            matrix_store (catwalk.storage.MatrixStore) a wrapper for the
                prediction matrix and metadata
            model_id (int) The database identifier of the model
        """
        labels = matrix_store.labels()
        matrix_type = matrix_store.matrix_type.string_name
        evaluation_start_time = matrix_store.as_of_dates[0]
        evaluation_end_time = matrix_store.as_of_dates[-1]
        as_of_date_frequency = matrix_store.metadata['as_of_date_frequency']

        # Specifies which evaluation table to write to: TestEvaluation or TrainEvaluation
        evaluation_table_obj = matrix_store.matrix_type.evaluation_obj

        logging.info(
            'Generating evaluations for model id %s, evaluation range %s-%s, '
            'as_of_date frequency %s',
            model_id,
            evaluation_start_time,
            evaluation_end_time,
            as_of_date_frequency
        )
        predictions_proba_sorted, labels_sorted = sort_predictions_and_labels(
            predictions_proba,
            labels,
            self.sort_seed
        )

        evaluations = []
        matrix_type = matrix_store.matrix_type
        if matrix_type.is_test:
            metric_groups_to_compute = self.testing_metric_groups
        else:
            metric_groups_to_compute = self.training_metric_groups
        for group in metric_groups_to_compute:
            evaluations = evaluations + self._evaluations_for_group(
                group,
                predictions_proba_sorted,
                labels_sorted,
                matrix_type.evaluation_obj
            )

        logging.info('Writing metrics to db: %s table', matrix_type)
        self._write_to_db(
            model_id,
            evaluation_start_time,
            evaluation_end_time,
            as_of_date_frequency,
            evaluations,
            evaluation_table_obj
        )
        logging.info('Done writing metrics to db: %s table', matrix_type)

    @db_retry
    def _write_to_db(
        self,
        model_id,
        evaluation_start_time,
        evaluation_end_time,
        as_of_date_frequency,
        evaluations,
        evaluation_table_obj
    ):
        """Write evaluation objects to the database

        Binds the model_id as as_of_date to the given ORM objects
        and writes them to the database

        Args:
            model_id (int) primary key of the model
            as_of_date (datetime.date) Date the predictions were made as of
            evaluations (list) results_schema.TestEvaluation or TrainEvaluation objects
            evaluation_table_obj (schema.TestEvaluation or TrainEvaluation)
                specifies to which table to add the evaluations
        """
        session = self.sessionmaker()

        session.query(evaluation_table_obj)\
            .filter_by(
                model_id=model_id,
                evaluation_start_time=evaluation_start_time,
                evaluation_end_time=evaluation_end_time,
                as_of_date_frequency=as_of_date_frequency
            ).delete()

        for evaluation in evaluations:
            evaluation.model_id = model_id
            evaluation.evaluation_start_time = evaluation_start_time
            evaluation.evaluation_end_time = evaluation_end_time
            evaluation.as_of_date_frequency = as_of_date_frequency
            session.add(evaluation)
        session.commit()
        session.close()



import logging
import operator

from triage.component.catwalk.evaluation import ModelEvaluator


def greater_is_better(metric):
    """Whether or not a metric wants higher values

    Args:
        metric (str): The name of a metric, ie 'precision@'
    Returns: (bool) Whether or not greater is better for the metric
    """
    if metric in ModelEvaluator.available_metrics:
        return ModelEvaluator.available_metrics[metric].greater_is_better
    else:
        logging.warning(
            'Metric %s not found in available metrics, assuming greater is better',
            metric
        )
        return True


def sql_rank_order(metric):
    """SQL Rank Order for a metric

    Args:
        metric (str): The name of a metric, ie 'precision@'
    Returns: (str) A SQL ORDER BY clause that will rank the best values first
    """
    if greater_is_better(metric):
        return 'desc'
    else:
        return 'asc'


def is_better_operator(metric):
    """Operator to decide which of two values is better

    Args:
        metric (str): The name of a metric, ie 'precision@'
    Returns: (function) An operator function that will compare two values
        and return whether or not the first one is better
    """
    if greater_is_better(metric):
        return operator.ge
    else:
        return operator.le


def best_in_series(metric):
    """The best value in a series

    Args:
        metric (str): The name of a metric, ie 'precision@'
    Returns: (str) The name of a pandas Series function that will provide
        the best value
    """
    if greater_is_better(metric):
        return 'max'
    else:
        return 'min'


def idxbest(metric):
    """Index of first occurrence of the best value

    Args:
        metric (str): The name of a metric, ie 'precision@'
    Returns: (str) The name of a pandas function that will provide
        the index of the first occurrence of the best value
    """
    if greater_is_better(metric):
        return 'idxmax'
    else:
        return 'idxmin'



import pandas
import testing.postgresql
import sqlalchemy
import unittest
from unittest.mock import patch
import pytest

from sqlalchemy import create_engine
from triage.component.catwalk.db import ensure_db
from tests.results_tests.factories import init_engine

from triage.component.catwalk.model_grouping import ModelGrouper
from triage.component.catwalk.model_trainers import ModelTrainer
from tests.utils import rig_engines, get_matrix_store


@pytest.fixture
def grid_config():
    return {
        'sklearn.tree.DecisionTreeClassifier': {
            'min_samples_split': [10, 100],
            'max_depth': [3, 5],
            'criterion': ['gini']
        }
    }


def test_model_trainer(grid_config):
    with rig_engines() as (db_engine, project_storage):
        # Creates a matrix entry in the matrices table with uuid from metadata above
        model_storage_engine = project_storage.model_storage_engine()
        trainer = ModelTrainer(
            experiment_hash=None,
            model_storage_engine=model_storage_engine,
            model_grouper=ModelGrouper(),
            db_engine=db_engine,
        )
        model_ids = trainer.train_models(
            grid_config=grid_config,
            misc_db_parameters=dict(),
            matrix_store=get_matrix_store(project_storage),
        )

        # assert
        # 1. that the models and feature importances table entries are present
        records = [
            row for row in
            db_engine.execute('select * from train_results.feature_importances')
        ]
        assert len(records) == 4 * 2  # maybe exclude entity_id? yes

        records = [
            row for row in
            db_engine.execute('select model_hash from model_metadata.models')
        ]
        assert len(records) == 4
        hashes = [row[0] for row in records]

        # 2. that the model groups are distinct
        records = [
            row for row in
            db_engine.execute('select distinct model_group_id from model_metadata.models')
        ]
        assert len(records) == 4

        # 3. that the model sizes are saved in the table and all are < 1 kB
        records = [
            row for row in
            db_engine.execute('select model_size from model_metadata.models')
        ]
        assert len(records) == 4
        for i in records:
            size = i[0]
            assert size < 1

        # 4. that all four models are cached
        model_pickles = [
            model_storage_engine.load(model_hash)
            for model_hash in hashes
        ]
        assert len(model_pickles) == 4
        assert len([x for x in model_pickles if x is not None]) == 4

        # 5. that their results can have predictions made on it
        test_matrix = pandas.DataFrame.from_dict({
            'entity_id': [3, 4],
            'feature_one': [4, 4],
            'feature_two': [6, 5],
        }).set_index('entity_id')


        for model_pickle in model_pickles:
            predictions = model_pickle.predict(test_matrix)
            assert len(predictions) == 2

        # 6. when run again, same models are returned
        new_model_ids = trainer.train_models(
            grid_config=grid_config,
            misc_db_parameters=dict(),
            matrix_store=get_matrix_store(project_storage)
        )
        assert len([
            row for row in
            db_engine.execute('select model_hash from model_metadata.models')
        ]) == 4
        assert model_ids == new_model_ids

        # 7. if replace is set, update non-unique attributes and feature importances
        max_batch_run_time = [
            row[0] for row in
            db_engine.execute('select max(batch_run_time) from model_metadata.models')
        ][0]
        trainer = ModelTrainer(
            experiment_hash=None,
            model_storage_engine=model_storage_engine,
            model_grouper=ModelGrouper(model_group_keys=['label_name', 'label_timespan']),
            db_engine=db_engine,
            replace=True
        )
        new_model_ids = trainer.train_models(
            grid_config=grid_config,
            misc_db_parameters=dict(),
            matrix_store=get_matrix_store(project_storage)
        )
        assert model_ids == new_model_ids
        assert [
            row['model_id'] for row in
            db_engine.execute('select model_id from model_metadata.models order by 1 asc')
        ] == model_ids
        new_max_batch_run_time = [
            row[0] for row in
            db_engine.execute('select max(batch_run_time) from model_metadata.models')
        ][0]
        assert new_max_batch_run_time > max_batch_run_time

        records = [
            row for row in
            db_engine.execute('select * from train_results.feature_importances')
        ]
        assert len(records) == 4 * 2  # maybe exclude entity_id? yes

        # 8. if the cache is missing but the metadata is still there, reuse the metadata
        for row in db_engine.execute('select model_hash from model_metadata.models'):
            model_storage_engine.delete(row[0])
        new_model_ids = trainer.train_models(
            grid_config=grid_config,
            misc_db_parameters=dict(),
            matrix_store=get_matrix_store(project_storage)
        )
        assert model_ids == sorted(new_model_ids)

        # 9. that the generator interface works the same way
        new_model_ids = trainer.generate_trained_models(
            grid_config=grid_config,
            misc_db_parameters=dict(),
            matrix_store=get_matrix_store(project_storage)
        )
        assert model_ids == \
            sorted([model_id for model_id in new_model_ids])

def test_baseline_exception_handling():
    grid_config = {
        'triage.component.catwalk.baselines.rankers.PercentileRankOneFeature': {
            'feature': ['feature_one', 'feature_three']
        }
    }
    with rig_engines() as (db_engine, project_storage):
        trainer = ModelTrainer(
            experiment_hash=None,
            model_storage_engine=project_storage.model_storage_engine(),
            db_engine=db_engine,
            model_grouper=ModelGrouper()
        )

        train_tasks = trainer.generate_train_tasks(
            grid_config,
            dict(),
            get_matrix_store(project_storage)
        )

        model_ids = []
        for train_task in train_tasks:
            model_ids.append(trainer.process_train_task(**train_task))
        assert model_ids == [1, None]


def test_custom_groups(grid_config):
    with rig_engines() as (db_engine, project_storage):
        # create training set
        model_storage_engine = project_storage.model_storage_engine()
        trainer = ModelTrainer(
            experiment_hash=None,
            model_storage_engine=model_storage_engine,
            model_grouper=ModelGrouper(['class_path']),
            db_engine=db_engine,
        )
        model_ids = trainer.train_models(
            grid_config=grid_config,
            misc_db_parameters=dict(),
            matrix_store=get_matrix_store(project_storage)
        )
        # expect only one model group now
        records = [
            row[0] for row in
            db_engine.execute('select distinct model_group_id from model_metadata.models')
        ]
        assert len(records) == 1
        assert records[0] == model_ids[0]


def test_n_jobs_not_new_model():
    grid_config = {
        'sklearn.ensemble.AdaBoostClassifier': {
            'n_estimators': [10, 100, 1000]
        },
        'sklearn.ensemble.RandomForestClassifier': {
            'n_estimators': [10, 100],
            'max_features': ['sqrt', 'log2'],
            'max_depth': [5, 10, 15, 20],
            'criterion': ['gini', 'entropy'],
            'n_jobs': [12, 24],
        }
    }

    with rig_engines() as (db_engine, project_storage):
        model_storage_engine = project_storage.model_storage_engine()
        trainer = ModelTrainer(
            experiment_hash=None,
            model_storage_engine=model_storage_engine,
            db_engine=db_engine,
            model_grouper=ModelGrouper()
        )

        train_tasks = trainer.generate_train_tasks(
            grid_config,
            dict(),
            get_matrix_store(project_storage),
        )

        assert len(train_tasks) == 35 # 32+3, would be (32*2)+3 if we didn't remove
        assert len([
            task for task in train_tasks
            if 'n_jobs' in task['parameters']
        ]) == 32

        for train_task in train_tasks:
            trainer.process_train_task(**train_task)

        for row in db_engine.execute(
            'select hyperparameters from model_metadata.model_groups'
        ):
            assert 'n_jobs' not in row[0]


class RetryTest(unittest.TestCase):
    def test_retry_max(self):
        db_engine = None
        trainer = None
        # set up a basic model training run
        with rig_engines() as (db_engine, project_storage):
            trainer = ModelTrainer(
                experiment_hash=None,
                model_storage_engine=project_storage.model_storage_engine(),
                db_engine=db_engine,
                model_grouper=ModelGrouper()
            )
            matrix_store = get_matrix_store(project_storage)

        # the postgres server goes out of scope here and thus no longer exists
        with patch('time.sleep') as time_mock:
            with self.assertRaises(sqlalchemy.exc.OperationalError):
                trainer.train_models(grid_config(), dict(), matrix_store)
            # we want to make sure that we are using the retrying module sanely
            # as opposed to matching the exact # of calls specified by the code
            assert len(time_mock.mock_calls) > 5

    def test_retry_recovery(self):
        db_engine = None
        trainer = None
        port = None
        with rig_engines() as (db_engine, project_storage):
            port = db_engine.url.port
            trainer = ModelTrainer(
                experiment_hash=None,
                model_storage_engine=project_storage.model_storage_engine(),
                db_engine=db_engine,
                model_grouper=ModelGrouper()
            )
            matrix_store = get_matrix_store(project_storage)

        # start without a database server
        # then bring it back up after the first sleep
        # use self so it doesn't go out of scope too early and shut down
        self.new_server = None

        def replace_db(arg):
            self.new_server = testing.postgresql.Postgresql(port=port)
            db_engine = create_engine(self.new_server.url())
            ensure_db(db_engine)
            init_engine(db_engine)
            get_matrix_store(project_storage)

        with patch('time.sleep') as time_mock:
            time_mock.side_effect = replace_db
            try:
                trainer.train_models(grid_config(), dict(), matrix_store)
            finally:
                if self.new_server is not None:
                    self.new_server.stop()
            assert len(time_mock.mock_calls) == 1



import copy
from datetime import date
from unittest import TestCase

import pandas
import pytest
import sqlalchemy
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.architect.feature_generators import FeatureGenerator
from triage.component.collate import (
    Aggregate,
    Categorical,
    SpacetimeAggregation,
)


INPUT_DATA = [
    # entity_id, knowledge_date, zip_code, cat_one, quantity_one
    (1, date(2014, 1, 1), '60120', 'good', 10000),
    (1, date(2014, 10, 11), '60120', 'good', None),
    (3, date(2012, 6, 8), '60653', 'bad', 342),
    (3, date(2014, 12, 21), '60653', 'inbetween', 600),
    (4, date(2014, 4, 4), '60653', 'bad', 1236)
]

INPUT_STATES = [
    # entity_id, as_of_date
    (1, date(2013, 9, 30)),
    (1, date(2014, 9, 30)),
    (1, date(2015, 1, 1)),
    (3, date(2013, 9, 30)),
    (3, date(2014, 9, 30)),
    (3, date(2015, 1, 1)),
    (4, date(2014, 9, 30)),
    (4, date(2015, 1, 1))
]


def setup_db(engine):
    engine.execute("""
        create table data (
            entity_id int,
            knowledge_date date,
            zip_code text,
            cat_one varchar,
            quantity_one float
        )
    """)
    for row in INPUT_DATA:
        engine.execute(
            'insert into data values (%s, %s, %s, %s, %s)',
            row
        )

    engine.execute("""
        create table states (
            entity_id int,
            as_of_date date
        )
    """)
    for row in INPUT_STATES:
        engine.execute(
            'insert into states values (%s, %s)',
            row
        )


def test_feature_generation():
    aggregate_config = [{
        'prefix': 'aprefix',
        'aggregates': [
            {
            'quantity': 'quantity_one',
            'metrics': ['sum', 'count'],
            'imputation': {
                'sum': {'type': 'constant', 'value': 137},
                'count': {'type': 'zero'}
                }
            },
        ],
        'categoricals_imputation': {
            'all': {'type': 'null_category'}
        },
        'categoricals': [
            {
                'column': 'cat_one',
                'choices': ['good', 'bad'],
                'metrics': ['sum']
            },
        ],
        'groups': ['entity_id', 'zip_code'],
        'intervals': ['all'],
        'knowledge_date_column': 'knowledge_date',
        'from_obj': 'data'
    }]

    expected_output = {
        'aprefix_aggregation_imputed': [
            {
                'entity_id': 1,
                'as_of_date': date(2013, 9, 30),
                'zip_code': None,
                'aprefix_entity_id_all_quantity_one_sum': 137,
                'aprefix_entity_id_all_quantity_one_count': 0,
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 0,
                'aprefix_entity_id_all_cat_one__NULL_sum': 1,
                'aprefix_zip_code_all_quantity_one_sum': 137,
                'aprefix_zip_code_all_quantity_one_count': 0,
                'aprefix_zip_code_all_cat_one_good_sum': 0,
                'aprefix_zip_code_all_cat_one_bad_sum': 0,
                'aprefix_zip_code_all_cat_one__NULL_sum': 1,
                'aprefix_entity_id_all_quantity_one_sum_imp': 1,
                'aprefix_entity_id_all_quantity_one_count_imp': 1,
                'aprefix_zip_code_all_quantity_one_sum_imp': 1,
                'aprefix_zip_code_all_quantity_one_count_imp': 1
            },
            {
                'entity_id': 1,
                'as_of_date': date(2014, 9, 30),
                'zip_code': '60120',
                'aprefix_entity_id_all_quantity_one_sum': 10000,
                'aprefix_entity_id_all_quantity_one_count': 1,
                'aprefix_entity_id_all_cat_one_good_sum': 1,
                'aprefix_entity_id_all_cat_one_bad_sum': 0,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0,
                'aprefix_zip_code_all_quantity_one_sum': 10000,
                'aprefix_zip_code_all_quantity_one_count': 1,
                'aprefix_zip_code_all_cat_one_good_sum': 1,
                'aprefix_zip_code_all_cat_one_bad_sum': 0,
                'aprefix_zip_code_all_cat_one__NULL_sum': 0,
                'aprefix_entity_id_all_quantity_one_sum_imp': 0,
                'aprefix_entity_id_all_quantity_one_count_imp': 0,
                'aprefix_zip_code_all_quantity_one_sum_imp': 0,
                'aprefix_zip_code_all_quantity_one_count_imp': 0
            },
            {
                'entity_id': 3,
                'as_of_date': date(2013, 9, 30),
                'zip_code': '60653',
                'aprefix_entity_id_all_quantity_one_sum': 342,
                'aprefix_entity_id_all_quantity_one_count': 1,
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0,
                'aprefix_zip_code_all_quantity_one_sum': 342,
                'aprefix_zip_code_all_quantity_one_count': 1,
                'aprefix_zip_code_all_cat_one_good_sum': 0,
                'aprefix_zip_code_all_cat_one_bad_sum': 1,
                'aprefix_zip_code_all_cat_one__NULL_sum': 0,
                'aprefix_entity_id_all_quantity_one_sum_imp': 0,
                'aprefix_entity_id_all_quantity_one_count_imp': 0,
                'aprefix_zip_code_all_quantity_one_sum_imp': 0,
                'aprefix_zip_code_all_quantity_one_count_imp': 0
            },
            {
                'entity_id': 3,
                'as_of_date': date(2014, 9, 30),
                'zip_code': '60653',
                'aprefix_entity_id_all_quantity_one_sum': 342,
                'aprefix_entity_id_all_quantity_one_count': 1,
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0,
                'aprefix_zip_code_all_quantity_one_sum': 1578,
                'aprefix_zip_code_all_quantity_one_count': 2,
                'aprefix_zip_code_all_cat_one_good_sum': 0,
                'aprefix_zip_code_all_cat_one_bad_sum': 2,
                'aprefix_zip_code_all_cat_one__NULL_sum': 0,
                'aprefix_entity_id_all_quantity_one_sum_imp': 0,
                'aprefix_entity_id_all_quantity_one_count_imp': 0,
                'aprefix_zip_code_all_quantity_one_sum_imp': 0,
                'aprefix_zip_code_all_quantity_one_count_imp': 0
            },
            {
                'entity_id': 4,
                'as_of_date': date(2014, 9, 30),
                'zip_code': '60653',
                'aprefix_entity_id_all_quantity_one_sum': 1236,
                'aprefix_entity_id_all_quantity_one_count': 1,
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0,
                'aprefix_zip_code_all_quantity_one_sum': 1578,
                'aprefix_zip_code_all_quantity_one_count': 2,
                'aprefix_zip_code_all_cat_one_good_sum': 0,
                'aprefix_zip_code_all_cat_one_bad_sum': 2,
                'aprefix_zip_code_all_cat_one__NULL_sum': 0,
                'aprefix_entity_id_all_quantity_one_sum_imp': 0,
                'aprefix_entity_id_all_quantity_one_count_imp': 0,
                'aprefix_zip_code_all_quantity_one_sum_imp': 0,
                'aprefix_zip_code_all_quantity_one_count_imp': 0
            },
        ]
    }

    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        setup_db(engine)

        features_schema_name = 'features'
        output_tables = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name
        ).create_all_tables(
            feature_dates=['2013-09-30', '2014-09-30'],
            feature_aggregation_config=aggregate_config,
            state_table='states',
        )

        for output_table in output_tables:
            records = pandas.read_sql(
                'select * from {}.{} order by entity_id, as_of_date'
                .format(features_schema_name, output_table),
                engine
            ).to_dict('records')
            for record, expected_record in zip(records, expected_output[output_table]):
                assert record == expected_record


def test_index_column_lookup():
    aggregations = [
        SpacetimeAggregation(
            prefix='prefix1',
            aggregates=[Categorical(
                col='cat_one',
                function='sum',
                choices=['good', 'bad', 'inbetween'],
                impute_rules={
                        'coltype': 'categorical',
                        'all': {'type': 'zero'}
                    }
            )],
            groups=['entity_id'],
            intervals=['all'],
            date_column='knowledge_date',
            output_date_column='as_of_date',
            dates=['2013-09-30', '2014-09-30'],
            state_table='states',
            state_group='entity_id',
            schema='features',
            from_obj='data'
        ),
        SpacetimeAggregation(
            prefix='prefix2',
            aggregates=[Aggregate(
                quantity='quantity_one',
                function='count',
                impute_rules={
                    'coltype': 'aggregate',
                    'all': {'type': 'zero'}
                }
            )],
            groups=['entity_id', 'zip_code'],
            intervals=['all'],
            date_column='knowledge_date',
            output_date_column='as_of_date',
            dates=['2013-09-30', '2014-09-30'],
            state_table='states',
            state_group='entity_id',
            schema='features',
            from_obj='data'
        )
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        setup_db(engine)

        features_schema_name = 'features'
        feature_generator = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name
        )
        lookup = feature_generator.index_column_lookup(aggregations)
        assert lookup == {
            'prefix1_aggregation_imputed': ['as_of_date', 'entity_id',],
            'prefix2_aggregation_imputed': ['as_of_date', 'entity_id', 'zip_code']
        }


def test_feature_generation_feature_start_time():
    aggregate_config = [{
        'prefix': 'aprefix',
        'aggregates_imputation': {'all': {'type': 'constant', 'value': 7}},
        'aggregates': [
            {'quantity': 'quantity_one', 'metrics': ['sum']},
        ],
        'groups': ['entity_id'],
        'intervals': ['all'],
        'knowledge_date_column': 'knowledge_date',
        'from_obj': 'data'
    }]

    expected_output = {
        'aprefix_aggregation_imputed': [
            {
                'entity_id': 1,
                'as_of_date': date(2015, 1, 1),
                'aprefix_entity_id_all_quantity_one_sum': 10000,
            },
            {
                'entity_id': 3,
                'as_of_date': date(2015, 1, 1),
                'aprefix_entity_id_all_quantity_one_sum': 600,
            },
            {
                'entity_id': 4,
                'as_of_date': date(2015, 1, 1),
                'aprefix_entity_id_all_quantity_one_sum': 1236,
            },
        ]
    }

    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        setup_db(engine)

        features_schema_name = 'features'
        output_tables = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name,
            feature_start_time='2013-01-01',
        ).create_all_tables(
            feature_dates=['2015-01-01'],
            feature_aggregation_config=aggregate_config,
            state_table='states',
        )

        for output_table in output_tables:
            records = pandas.read_sql(
                'select * from {}.{} order by as_of_date, entity_id'
                .format(features_schema_name, output_table),
                engine
            ).to_dict('records')
            assert records == expected_output[output_table]


def test_dynamic_categoricals():
    aggregate_config = [{
        'prefix': 'aprefix',
        'categoricals': [
            {
                'column': 'cat_one',
                'choice_query': 'select distinct(cat_one) from data',
                'metrics': ['sum'],
                'imputation': {'all': {'type': 'null_category'}}
            },
        ],
        'groups': ['entity_id'],
        'intervals': ['all'],
        'knowledge_date_column': 'knowledge_date',
        'from_obj': 'data'
    }]
    expected_output = {
        'aprefix_aggregation_imputed': [
            {
                'entity_id': 1,
                'as_of_date': date(2013, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 0,
                'aprefix_entity_id_all_cat_one__NULL_sum': 1
            },
            {
                'entity_id': 3,
                'as_of_date': date(2013, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0
            },
            {
                'entity_id': 1,
                'as_of_date': date(2014, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 1,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 0,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0
            },
            {
                'entity_id': 3,
                'as_of_date': date(2014, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0
            },
            {
                'entity_id': 4,
                'as_of_date': date(2014, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0
            },
        ]
    }

    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        setup_db(engine)

        features_schema_name = 'features'

        output_tables = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name
        ).create_all_tables(
            feature_dates=['2013-09-30', '2014-09-30'],
            feature_aggregation_config=aggregate_config,
            state_table = 'states'
        )

        for output_table in output_tables:
            records = pandas.read_sql(
                'select * from {}.{} order by as_of_date, entity_id'.format(
                    features_schema_name, output_table
                ),
                engine
            ).to_dict('records')
            assert records == expected_output[output_table]


def test_array_categoricals():
    aggregate_config = [{
        'prefix': 'aprefix',
        'array_categoricals': [
            {
                'column': 'cat_one',
                'choices': ['good', 'bad', 'inbetween'],
                'metrics': ['sum'],
                'imputation': {'all': {'type': 'null_category'}}
            },
        ],
        'groups': ['entity_id'],
        'intervals': ['all'],
        'knowledge_date_column': 'knowledge_date',
        'from_obj': 'data'
    }]
    expected_output = {
        'aprefix_aggregation_imputed': [
            {
                'entity_id': 1,
                'as_of_date': date(2013, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 0,
                'aprefix_entity_id_all_cat_one__NULL_sum': 1
            },
            {
                'entity_id': 3,
                'as_of_date': date(2013, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0
            },
            {
                'entity_id': 1,
                'as_of_date': date(2014, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 1,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 0,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0
            },
            {
                'entity_id': 3,
                'as_of_date': date(2014, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0
            },
            {
                'entity_id': 4,
                'as_of_date': date(2014, 9, 30),
                'aprefix_entity_id_all_cat_one_good_sum': 0,
                'aprefix_entity_id_all_cat_one_inbetween_sum': 0,
                'aprefix_entity_id_all_cat_one_bad_sum': 1,
                'aprefix_entity_id_all_cat_one__NULL_sum': 0
            },
        ]
    }

    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        input_data = [
            # entity_id, knowledge_date, cat_one, quantity_one
            (1, date(2014, 1, 1), ['good', 'good'], 10000),
            (1, date(2014, 10, 11), ['good'], None),
            (3, date(2012, 6, 8), ['bad'], 342),
            (3, date(2014, 12, 21), ['inbetween'], 600),
            (4, date(2014, 4, 4), ['bad'], 1236)
        ]

        engine.execute("""
            create table data (
                entity_id int,
                knowledge_date date,
                cat_one varchar[],
                quantity_one float
            )
        """)
        for row in input_data:
            engine.execute(
                'insert into data values (%s, %s, %s, %s)',
                row
            )

        engine.execute("""
            create table states (
                entity_id int,
                as_of_date date
            )
        """)
        for row in INPUT_STATES:
            engine.execute(
                'insert into states values (%s, %s)',
                row
            )

        features_schema_name = 'features'

        output_tables = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name
        ).create_all_tables(
            feature_dates=['2013-09-30', '2014-09-30'],
            feature_aggregation_config=aggregate_config,
            state_table = 'states'
        )

        for output_table in output_tables:
            records = pandas.read_sql(
                'select * from {}.{} order by as_of_date, entity_id'.format(
                    features_schema_name, output_table
                ),
                engine
            ).to_dict('records')
            assert records == expected_output[output_table]


def test_generate_table_tasks():
    aggregations = [
        SpacetimeAggregation(
            prefix='prefix1',
            aggregates=[Categorical(
                col='cat_one',
                function='sum',
                choices=['good', 'bad', 'inbetween'],
                impute_rules={
                        'coltype': 'categorical',
                        'all': {'type': 'zero'}
                    }
            )],
            groups=['entity_id'],
            intervals=['all'],
            date_column='knowledge_date',
            output_date_column='as_of_date',
            dates=['2013-09-30', '2014-09-30'],
            state_table = 'states',
            state_group = 'entity_id',
            schema='features',
            from_obj='data'
        ),
        SpacetimeAggregation(
            prefix='prefix2',
            aggregates=[Aggregate(
                quantity='quantity_one',
                function='count',
                impute_rules={
                    'coltype': 'aggregate',
                    'all': {'type': 'zero'}
                }
            )],
            groups=['entity_id'],
            intervals=['all'],
            date_column='knowledge_date',
            output_date_column='as_of_date',
            dates=['2013-09-30', '2014-09-30'],
            state_table = 'states',
            state_group = 'entity_id',
            schema='features',
            from_obj='data'
        )
    ]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        setup_db(engine)

        features_schema_name = 'features'

        table_tasks = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name
        ).generate_all_table_tasks(aggregations, task_type='aggregation')
        for table_name, task in table_tasks.items():
            assert 'DROP TABLE' in task['prepare'][0]
            assert 'CREATE TABLE' in str(task['prepare'][1])
            assert 'CREATE INDEX' in task['finalize'][0]
            assert isinstance(task['inserts'], list)

        # build the aggregation tables to check the imputation tasks
        FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name
        ).process_table_tasks(table_tasks)

        table_tasks = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name
        ).generate_all_table_tasks(aggregations, task_type='imputation')
        for table_name, task in table_tasks.items():
            assert 'DROP TABLE' in task['prepare'][0]
            assert 'CREATE TABLE' in str(task['prepare'][1])
            assert 'CREATE INDEX' in task['finalize'][0]
            assert isinstance(task['inserts'], list)


def test_aggregations():
    aggregate_config = [{
        'prefix': 'prefix1',
        'categoricals': [
            {
                'column': 'cat_one',
                'choice_query': 'select distinct(cat_one) from data',
                'metrics': ['sum'],
                'imputation': {'all': {'type': 'null_category'}}
            },
        ],
        'groups': ['entity_id'],
        'intervals': ['all'],
        'knowledge_date_column': 'knowledge_date',
        'from_obj': 'data'
    }, {
        'prefix': 'prefix2',
        'aggregates_imputation': {'all': {'type': 'mean'}},
        'aggregates': [
            {'quantity': 'quantity_one', 'metrics': ['count']},
        ],
        'groups': ['entity_id'],
        'intervals': ['all'],
        'knowledge_date_column': 'knowledge_date',
        'from_obj': 'data'
    }]
    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        setup_db(engine)

        features_schema_name = 'features'

        aggregations = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name
        ).aggregations(
            feature_dates=['2013-09-30', '2014-09-30'],
            feature_aggregation_config=aggregate_config,
            state_table='states'
        )
        for aggregation in aggregations:
            assert isinstance(aggregation, SpacetimeAggregation)

def test_replace():
    aggregate_config = [{
        'prefix': 'aprefix',
        'aggregates_imputation': {'all': {'type': 'mean'}},
        'aggregates': [
            {'quantity': 'quantity_one', 'metrics': ['sum', 'count']},
        ],
        'categoricals': [
            {
                'column': 'cat_one',
                'choices': ['good', 'bad'],
                'metrics': ['sum'],
                'imputation': {'all': {'type': 'null_category'}}
            },
        ],
        'groups': ['entity_id'],
        'intervals': ['all'],
        'knowledge_date_column': 'knowledge_date',
        'from_obj': 'data'
    }]

    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        setup_db(engine)

        features_schema_name = 'features'
        feature_tables = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name,
            replace=False
        ).create_all_tables(
            feature_dates=['2013-09-30', '2014-09-30'],
            feature_aggregation_config=aggregate_config,
            state_table='states'
        )

        assert len(feature_tables) == 1
        assert list(feature_tables)[0] == 'aprefix_aggregation_imputed'

        feature_generator = FeatureGenerator(
            db_engine=engine,
            features_schema_name=features_schema_name,
            replace=False
        )
        aggregations = feature_generator.aggregations(
            feature_dates=['2013-09-30', '2014-09-30'],
            feature_aggregation_config=aggregate_config,
            state_table='states'
        )
        table_tasks = feature_generator.generate_all_table_tasks(aggregations, task_type='aggregation')

        assert len(table_tasks['aprefix_entity_id'].keys()) == 0

        imp_tasks = feature_generator.generate_all_table_tasks(aggregations, task_type='imputation')

        assert len(imp_tasks['aprefix_aggregation_imputed'].keys()) == 0

        engine.dispose()


def test_transaction_error():
    """Database connections are cleaned up regardless of in-transaction
    query errors.

    """
    aggregate_config = [{
        'prefix': 'aprefix',
        'aggregates': [
            {
                'quantity': 'quantity_one',
                'metrics': ['sum'],
                'imputation': {
                    'sum': {'type': 'constant', 'value': 137},
                    'count': {'type': 'zero'}
                }
            },
        ],
        'groups': ['entity_id'],
        'intervals': ['all'],
        'knowledge_date_column': 'knowledge_date',
        'from_obj': 'data'
    }]

    with testing.postgresql.Postgresql() as postgresql:
        engine = create_engine(postgresql.url())
        setup_db(engine)

        feature_generator = FeatureGenerator(db_engine=engine,
                                             features_schema_name='features')
        with pytest.raises(sqlalchemy.exc.ProgrammingError):
            feature_generator.create_all_tables(
                feature_dates=['2013-09-30', '2014-09-30'],
                feature_aggregation_config=aggregate_config,
                state_table='statez',  # WRONG!
            )

        (query_count,) = engine.execute('''\
            select count(*) from pg_stat_activity
            where query not ilike '%%pg_stat_activity%%'
        ''').fetchone()

        assert query_count == 0

        engine.dispose()


class TestValidations(TestCase):

    def setUp(self):
        self.postgresql = testing.postgresql.Postgresql()
        engine = create_engine(self.postgresql.url())
        setup_db(engine)
        self.feature_generator = FeatureGenerator(engine, 'features')

        self.base_config = {
            'prefix': 'aprefix',
            'categoricals': [
                {
                    'column': 'cat_one',
                    'choices': ['good', 'bad'],
                    'metrics': ['sum'],
                    'imputation': {'all': {'type': 'null_category'}}
                },
            ],
            'groups': ['entity_id', 'zip_code'],
            'intervals': ['all'],
            'knowledge_date_column': 'knowledge_date',
            'from_obj': 'data'
        }

    def tearDown(self):
        self.postgresql.stop()

    def test_correct_keys(self):
        self.feature_generator.validate([self.base_config])

        with self.assertRaises(ValueError):
            no_group = copy.deepcopy(self.base_config)
            del no_group['groups']
            self.feature_generator.validate([no_group])

        with self.assertRaises(ValueError):
            no_intervals = copy.deepcopy(self.base_config)
            del no_intervals['intervals']
            self.feature_generator.validate([no_intervals])

        with self.assertRaises(ValueError):
            no_kdate = copy.deepcopy(self.base_config)
            del no_kdate['knowledge_date_column']
            self.feature_generator.validate([no_kdate])

        with self.assertRaises(ValueError):
            no_from_obj = copy.deepcopy(self.base_config)
            del no_from_obj['from_obj']
            self.feature_generator.validate([no_from_obj])

        with self.assertRaises(ValueError):
            no_aggs = copy.deepcopy(self.base_config)
            del no_aggs['categoricals']
            self.feature_generator.validate([no_aggs])

        with self.assertRaises(ValueError):
            no_imps = copy.deepcopy(self.base_config)
            del no_imps['categoricals'][0]['imputation']
            self.feature_generator.validate([no_imps])

    def test_bad_from_obj(self):
        bad_from_obj = copy.deepcopy(self.base_config)
        bad_from_obj['from_obj'] = 'where thing is other_thing'
        with self.assertRaises(ValueError):
            self.feature_generator.validate([bad_from_obj])

    def test_bad_interval(self):
        bad_interval = copy.deepcopy(self.base_config)
        bad_interval['intervals'] = ['1y', '1fortnight']
        with self.assertRaises(ValueError):
            self.feature_generator.validate([bad_interval])

    def test_bad_group(self):
        bad_group = copy.deepcopy(self.base_config)
        bad_group['groups'] = ['zip_code', 'otherthing']
        with self.assertRaises(ValueError):
            self.feature_generator.validate([bad_group])

    def test_bad_choice_query(self):
        bad_choice_query = copy.deepcopy(self.base_config)
        del bad_choice_query['categoricals'][0]['choices']
        bad_choice_query['categoricals'][0]['choice_query'] = 'select distinct cat_two from data'
        with self.assertRaises(ValueError):
            self.feature_generator.validate([bad_choice_query])

    def test_wrong_imp_fcn(self):
        wrong_imp_fcn = copy.deepcopy(self.base_config)
        del wrong_imp_fcn['categoricals'][0]['imputation']['all']
        wrong_imp_fcn['categoricals'][0]['imputation']['max'] = {'type': 'null_category'}
        with self.assertRaises(ValueError):
            self.feature_generator.validate([wrong_imp_fcn])

    def test_bad_imp_rule(self):
        bad_imp_rule = copy.deepcopy(self.base_config)
        bad_imp_rule['categoricals'][0]['imputation']['all'] = {'type': 'bad_rule_doesnt_exist'}
        with self.assertRaises(ValueError):
            self.feature_generator.validate([bad_imp_rule])

    def test_no_imp_rule_type(self):
        no_imp_rule_type = copy.deepcopy(self.base_config)
        no_imp_rule_type['categoricals'][0]['imputation']['all'] = {'value': 'good'}
        with self.assertRaises(ValueError):
            self.feature_generator.validate([no_imp_rule_type])

    def test_missing_imp_arg(self):
        missing_imp_arg = copy.deepcopy(self.base_config)
        # constant value imputation requires a 'value' parameter
        missing_imp_arg['categoricals'][0]['imputation']['all'] = {'type': 'constant'}
        with self.assertRaises(ValueError):
            self.feature_generator.validate([missing_imp_arg])



import numpy as np
import pandas as pd

import pytest
from unittest.mock import MagicMock
from unittest import TestCase

from triage.component.catwalk.baselines.rankers import PercentileRankOneFeature
from triage.component.catwalk.baselines.thresholders import SimpleThresholder
from triage.component.catwalk.exceptions import BaselineFeatureNotInMatrix


@pytest.fixture(scope='class')
def data(request):
    X_train = pd.DataFrame({
        'x1': [0,  1, 2, 56, 25, 8, -3, 89],
        'x2': [0, 23, 1,  6,  5, 3, 18,  7],
        'x3': [1, 12, 5, -6,  2, 5,  3, -3],
        'x4': [6, 13, 4,  5, 35, 6, 43, 74]
    })
    y_train = [0,  1, 0,  1,  1, 1,  3,  0]
    X_test = pd.DataFrame({
        'x1': [4, 14, 0,  6, 25, 8, -3,  4],
        'x2': [6, -1, 1, 24,  5, 3, 18, 39],
        'x3': [1,  7, 4, 57,  2, 5,  3,  2],
        'x4': [7,  3, 6, 39, 35, 6, 43, -6]
    })
    y_test  = [1,  3, 0,  0,  0, 0,  0,  1]

    request.cls.data = {'X_train':X_train, 'X_test':X_test, 'y_train':y_train, 'y_test':y_test}

@pytest.fixture(scope='class')
def rules(request):
    request.cls.rules = [
        'x1 > 0',
        'x2 <= 1'
    ]

@pytest.mark.usefixtures('data')
class TestRankOneFeature(TestCase):
    def test_fit(self):
        ranker = PercentileRankOneFeature(feature='x3')
        assert ranker.feature_importances_ == None
        ranker.fit(x=self.data['X_train'], y=self.data['y_train'])
        np.testing.assert_array_equal(ranker.feature_importances_, np.array([0, 0, 1, 0]))

    def test_ranking_on_unavailable_feature_raises_error(self):
        ranker = PercentileRankOneFeature(feature='x5')
        with self.assertRaises(BaselineFeatureNotInMatrix):
            ranker.fit(x=self.data['X_train'], y=self.data['y_train'])

    def test_predict_proba(self):
        for descend_value in [True, False]:
            ranker = PercentileRankOneFeature(feature='x3', descend=descend_value)
            ranker.fit(x=self.data['X_train'], y=self.data['y_train'])
            results = ranker.predict_proba(self.data['X_test'])
            if descend_value:
                expected_results = np.array(
                    [np.zeros(len(self.data['X_test'])),
                    [.875, .125, .375, 0, .625, .25, .5, .625]]
                ).transpose()
            else:
                expected_results = np.array(
                    [np.zeros(len(self.data['X_test'])),
                    [0,  .75, .5, .875, .125, .625, .375, .125]]
                ).transpose()
            np.testing.assert_array_equal(results, expected_results)


@pytest.mark.usefixtures('data', 'rules')
class TestSimpleThresholder(TestCase):
    def test__convert_string_rule_to_dict(self):
        thresholder = SimpleThresholder(self.rules, 'or')
        results = thresholder._convert_string_rule_to_dict(self.rules[0           ])
        expected_results = {
                'feature_name': 'x1',
                'operator': 'gt',
                'threshold': 0
            }
        assert results == expected_results

    def test_all_feature_names_property(self):
        thresholder = SimpleThresholder(self.rules, 'or')
        assert thresholder.all_feature_names == ['x1', 'x2']

    def test_fit(self):
        thresholder = SimpleThresholder(self.rules, 'or')
        assert thresholder.feature_importances_ == None
        thresholder.fit(x=self.data['X_train'], y=self.data['y_train'])
        np.testing.assert_array_equal(thresholder.feature_importances_, np.array([1, 1, 0, 0]))

    def test_rule_with_unavailable_feature_raises_error(self):
        rules = self.rules
        rules.append('x5 == 3')
        thresholder = SimpleThresholder(rules, 'or')
        with self.assertRaises(BaselineFeatureNotInMatrix):
            thresholder.fit(x=self.data['X_train'], y=self.data['y_train'])

    def test_predict_proba(self):
        for logical_operator in ['and', 'or']:
            thresholder = SimpleThresholder(self.rules, logical_operator)
            thresholder.fit(x=self.data['X_train'], y=self.data['y_train'])
            results = thresholder.predict_proba(self.data['X_test'])
            if logical_operator == 'and':
                expected_results = np.array(
                    [[0,  1, 0,  0,  0, 0,  0,  0], [0,  1, 0,  0,  0, 0,  0,  0]]
                ).transpose()
            elif logical_operator == 'or':
                expected_results = np.array(
                    [[1,  1, 1,  1,  1, 1,  0,  1], [1,  1, 1,  1,  1, 1,  0,  1]]
                ).transpose()
            np.testing.assert_array_equal(results, expected_results)



#!/usr/bin/env python
from datetime import datetime
import os
import yaml
import argparse
from descriptors import cachedproperty
from argcmdr import RootCommand, Command, main, cmdmethod
from sqlalchemy.engine.url import URL

import logging

from triage.component.architect.feature_generators import FeatureGenerator
from triage.component.audition import AuditionRunner
from triage.component.results_schema import upgrade_db, stamp_db, REVISION_MAPPING
from triage.component.timechop import Timechop
from triage.component.timechop.plotting import visualize_chops
from triage.experiments import CONFIG_VERSION, MultiCoreExperiment, SingleThreadedExperiment
from triage.util.db import create_engine

logging.basicConfig(level=logging.INFO)

import importlib.util


def natural_number(value):
    natural = int(value)
    if natural <= 0:
        raise argparse.ArgumentTypeError(f"{value} is an invalid natural number")
    return natural


def valid_date(value):
    try:
        return datetime.strptime(value, "%Y-%m-%d")
    except ValueError:
        raise argparse.ArgumentTypeError(f"{value} is an invalid date (format: YYYY-MM-DD)")


class Triage(RootCommand):
    """manage Triage database and experiments"""

    def __init__(self, parser):
        parser.add_argument(
            '-d', '--dbfile',
            default='database.yaml',
            type=argparse.FileType('r'),
            help="database connection file",
        )
        parser.add_argument(
            '-s', '--setup',
            help="file path to Python module to import before running the Experiment",
        )

    def setup(self):
        setup_path = self.args.setup or os.path.abspath('experiment.py')
        if setup_path is not None:
            logging.info(f"Loading configurations from {setup_path}")
            spec = importlib.util.spec_from_file_location("triage_config", setup_path)
            triage_config = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(triage_config)
            logging.info(f"Configuration loaded")
        return None

    @cachedproperty
    def db_url(self):
        dbconfig = yaml.load(self.args.dbfile)
        db_url = URL(
                'postgres',
                host=dbconfig['host'],
                username=dbconfig['user'],
                database=dbconfig['db'],
                password=dbconfig['pass'],
                port=dbconfig['port'],
            )
        return db_url

    @cmdmethod
    def configversion(self, args):
        """Check the experiment config version compatible with this installation of Triage"""
        print(CONFIG_VERSION)


@Triage.register
class ShowTimeChops(Command):
    """Visualize time chops (temporal cross-validation blocks')"""
    def __init__(self, parser):
        parser.add_argument(
            'config',
            type=argparse.FileType('r'),
            help="YAML file containing temporal_config (for instance, an Experiment config)"
        )

    def __call__(self, args):
        experiment_config = yaml.load(args.config)
        if 'temporal_config' not in experiment_config:
            raise ValueError('Passed configuration must have `temporal_config` key '
                             'in order to visualize time chops')
        chopper = Timechop(**(experiment_config['temporal_config']))
        logging.info('Visualizing time chops')
        visualize_chops(chopper)


@Triage.register
class FeatureTest(Command):
    """Test a feature aggregation by running it for one date"""
    def __init__(self, parser):
        parser.add_argument(
            'as_of_date',
            type=valid_date,
            help='The date as of which to run features. Format YYYY-MM-DD'
        )
        parser.add_argument(
            'feature_config_file',
            type=argparse.FileType('r'),
            help='Feature config YAML file, containing a list of feature_aggregation objects'
        )

    def __call__(self, args):
        self.root.setup  # Loading configuration (if exists)
        db_engine = create_engine(self.root.db_url)
        feature_config = yaml.load(args.feature_config_file)

        FeatureGenerator(db_engine, 'features_test').create_features_before_imputation(
            feature_aggregation_config=feature_config,
            feature_dates=[args.as_of_date]
        )
        logging.info('Features created for feature_config %s and date %s', feature_config, args.as_of_date)


@Triage.register
class Experiment(Command):
    """Run a full modeling experiment"""

    def __init__(self, parser):
        parser.add_argument(
            'config',
            type=argparse.FileType('r'),
            help="config file for Experiment"
        )
        parser.add_argument(
            '--project-path',
            default=os.path.curdir,
            help="path to store matrices and trained models"
        )
        parser.add_argument(
            '--n-db-processes',
            type=natural_number,
            default=1,
            help="number of concurrent database connections to use"
        )
        parser.add_argument(
            '--n-processes',
            type=natural_number,
            default=1,
            help="number of cores to use"
        )
        parser.add_argument(
            '--replace',
            dest='replace',
            action='store_true'
        )
        parser.add_argument(
            '-v', '--validate',
            action='store_true',
            help='validate before running experiment'
        )
        parser.add_argument(
            '--no-validate',
            action='store_false',
            dest='validate',
            help="run experiment without validation"
        )
        parser.add_argument(
            '--validate-only',
            action='store_true',
            help="only validate the config file not running Experiment"
        )

        parser.set_defaults(
            validate=True,
            validate_only=False,
        )

    @cachedproperty
    def experiment(self):
        self.root.setup  # Loading configuration (if exists)
        db_url = self.root.db_url
        config = yaml.load(self.args.config)
        db_engine = create_engine(db_url)
        common_kwargs = {
            'db_engine': db_engine,
            'project_path': self.args.project_path,
            'config': config,
            'replace': self.args.replace,
        }
        if self.args.n_db_processes > 1 or self.args.n_processes > 1:
            experiment = MultiCoreExperiment(
                n_db_processes=self.args.n_db_processes,
                n_processes=self.args.n_processes,
                **common_kwargs
            )
        else:
            experiment = SingleThreadedExperiment(**common_kwargs)
        return experiment

    def __call__(self, args):
        if args.validate_only:
            self.experiment.validate()
        elif args.validate:
            self.experiment.validate()
            self.experiment.run()
        else:
            self.experiment.run()


@Triage.register
class Audition(Command):
    """Audition models from a completed experiment to pick a smaller group of promising models
    """
    def __init__(self, parser):
        parser.add_argument(
            '-c', '--config',
            type=argparse.FileType('r'),
            default='audition_config.yaml',
            help="config file for audition",
        )
        parser.set_defaults(
            directory=None,
            validate=True,
        )

    @cachedproperty
    def runner(self):
        self.root.setup # Loading configuration (if exists)
        db_url = self.root.db_url
        dir_plot = self.args.directory
        config = yaml.load(self.args.config)
        db_engine = create_engine(db_url)
        return AuditionRunner(config, db_engine, dir_plot)

    def __call__(self, args):
        self['run'](args)

    @cmdmethod('-d', '--directory', default=None, help="directory to store the result plots from audition")
    @cmdmethod('-v', '--validate', action='store_true', help="validate before running audition")
    @cmdmethod('--no-validate', action='store_false', dest='validate', help="run audtion without validation")
    @cmdmethod('--validate-only', action='store_true', help="only validate the config file not running audition")
    def run(self, args):
        if args.validate_only:
            try:
                self.runner.validate()
            except Exception as err:
                raise(err)
        elif args.validate:
            try:
                self.runner.validate()
                self.runner.run()
            except Exception as err:
                raise(err)
        else:
            self.runner.run()


@Triage.register
class Db(Command):
    """Manage experiment database"""

    @cmdmethod
    def upgrade(self, args):
        """Upgrade triage results database"""
        upgrade_db(args.dbfile)

    @cmdmethod('configversion', choices=REVISION_MAPPING.keys(), help='config version of last experiment you ran')
    def stamp(self, args):
        """Instruct the triage results database to mark itself as updated to a known version without doing any upgrading.

        Use this if the database was created without an 'alembic_version' table. Uses the config version of your experiment to infer what database version is suitable.
        """
        revision = REVISION_MAPPING[args.configversion]
        print(f"Based on config version {args.configversion} "
              f"we think your results schema is version {revision} and are upgrading to it")
        stamp_db(revision, args.dbfile)


def execute():
    main(Triage)


if __name__ == '__main__':
    main(Triage)



"""empty message

Revision ID: 8b3f167d0418
Revises:
Create Date: 2017-05-11 12:03:37.001446

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '8b3f167d0418'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.execute('CREATE SCHEMA results')
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'experiments',
        sa.Column('experiment_hash', sa.String(), nullable=False),
        sa.Column('config', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.PrimaryKeyConstraint('experiment_hash'),
        schema='results'
    )
    op.create_table(
        'model_groups',
        sa.Column('model_group_id', sa.Integer(), nullable=False),
        sa.Column('model_type', sa.Text(), nullable=True),
        sa.Column('model_parameters', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column('feature_list', sa.ARRAY(sa.Text()), nullable=True),
        sa.Column('model_config', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.PrimaryKeyConstraint('model_group_id'),
        schema='results'
    )
    op.create_table(
        'models',
        sa.Column('model_id', sa.Integer(), nullable=False),
        sa.Column('model_group_id', sa.Integer(), nullable=True),
        sa.Column('model_hash', sa.String(), nullable=True),
        sa.Column('run_time', sa.DateTime(), nullable=True),
        sa.Column('batch_run_time', sa.DateTime(), nullable=True),
        sa.Column('model_type', sa.String(), nullable=True),
        sa.Column('model_parameters', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column('model_comment', sa.Text(), nullable=True),
        sa.Column('batch_comment', sa.Text(), nullable=True),
        sa.Column('config', sa.JSON(), nullable=True),
        sa.Column('experiment_hash', sa.String(), nullable=True),
        sa.Column('train_end_time', sa.DateTime(), nullable=True),
        sa.Column('test', sa.Boolean(), nullable=True),
        sa.Column('train_matrix_uuid', sa.Text(), nullable=True),
        sa.Column('train_label_window', sa.Interval(), nullable=True),
        sa.ForeignKeyConstraint(['experiment_hash'], ['results.experiments.experiment_hash'], ),
        sa.ForeignKeyConstraint(['model_group_id'], ['results.model_groups.model_group_id'], ),
        sa.PrimaryKeyConstraint('model_id'),
        schema='results'
    )
    op.create_index(
        op.f('ix_results_models_model_hash'),
        'models',
        ['model_hash'],
        unique=True,
        schema='results',
    )
    op.create_table(
        'evaluations',
        sa.Column('model_id', sa.Integer(), nullable=False),
        sa.Column('evaluation_start_time', sa.DateTime(), nullable=False),
        sa.Column('evaluation_end_time', sa.DateTime(), nullable=False),
        sa.Column('example_frequency', sa.Interval(), nullable=False),
        sa.Column('metric', sa.String(), nullable=False),
        sa.Column('parameter', sa.String(), nullable=False),
        sa.Column('value', sa.Numeric(), nullable=True),
        sa.Column('num_labeled_examples', sa.Integer(), nullable=True),
        sa.Column('num_labeled_above_threshold', sa.Integer(), nullable=True),
        sa.Column('num_positive_labels', sa.Integer(), nullable=True),
        sa.Column('sort_seed', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['model_id'], ['results.models.model_id'], ),
        sa.PrimaryKeyConstraint(
            'model_id',
            'evaluation_start_time',
            'evaluation_end_time',
            'example_frequency',
            'metric',
            'parameter',
        ),
        schema='results'
    )
    op.create_table(
        'feature_importances',
        sa.Column('model_id', sa.Integer(), nullable=False),
        sa.Column('feature', sa.String(), nullable=False),
        sa.Column('feature_importance', sa.Numeric(), nullable=True),
        sa.Column('rank_abs', sa.Integer(), nullable=True),
        sa.Column('rank_pct', sa.Float(), nullable=True),
        sa.ForeignKeyConstraint(['model_id'], ['results.models.model_id'], ),
        sa.PrimaryKeyConstraint('model_id', 'feature'),
        schema='results'
    )
    op.create_table(
        'predictions',
        sa.Column('model_id', sa.Integer(), nullable=False),
        sa.Column('entity_id', sa.BigInteger(), nullable=False),
        sa.Column('as_of_date', sa.DateTime(), nullable=False),
        sa.Column('score', sa.Numeric(), nullable=True),
        sa.Column('label_value', sa.Integer(), nullable=True),
        sa.Column('rank_abs', sa.Integer(), nullable=True),
        sa.Column('rank_pct', sa.Float(), nullable=True),
        sa.Column('matrix_uuid', sa.Text(), nullable=True),
        sa.Column('test_label_window', sa.Interval(), nullable=True),
        sa.ForeignKeyConstraint(['model_id'], ['results.models.model_id'], ),
        sa.PrimaryKeyConstraint('model_id', 'entity_id', 'as_of_date'),
        schema='results'
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('predictions', schema='results')
    op.drop_table('feature_importances', schema='results')
    op.drop_table('evaluations', schema='results')
    op.drop_index(op.f('ix_results_models_model_hash'), table_name='models', schema='results')
    op.drop_table('models', schema='results')
    op.drop_table('model_groups', schema='results')
    op.drop_table('experiments', schema='results')
    op.execute('DROP SCHEMA results')
    # ### end Alembic commands ###



from unittest import TestCase
from datetime import datetime
import testing.postgresql
from sqlalchemy import create_engine

from triage.component.audition.distance_from_best import DistanceFromBestTable
from triage.component.audition.thresholding import model_groups_filter, ModelGroupThresholder
from triage.component.catwalk.db import ensure_db

from tests.results_tests.factories import ModelFactory, ModelGroupFactory, init_engine, session

class ModelGroupFilterTest(TestCase):
    def filter_same_train_end_times(self, engine):
        ensure_db(engine)
        init_engine(engine)
        mg1 = ModelGroupFactory(model_group_id=1, model_type='modelType1')
        mg2 = ModelGroupFactory(model_group_id=2, model_type='modelType2')
        mg3 = ModelGroupFactory(model_group_id=3, model_type='modelType3')
        mg4 = ModelGroupFactory(model_group_id=4, model_type='modelType4')
        # model group 1
        ModelFactory(model_group_rel=mg1, train_end_time=datetime(2014, 1, 1))
        ModelFactory(model_group_rel=mg1, train_end_time=datetime(2015, 1, 1))
        ModelFactory(model_group_rel=mg1, train_end_time=datetime(2016, 1, 1))
        ModelFactory(model_group_rel=mg1, train_end_time=datetime(2017, 1, 1))
        # model group 2 only has three timestamps, should not pass
        ModelFactory(model_group_rel=mg2, train_end_time=datetime(2014, 1, 1))
        # model group 3
        ModelFactory(model_group_rel=mg3, train_end_time=datetime(2014, 1, 1))
        ModelFactory(model_group_rel=mg3, train_end_time=datetime(2015, 1, 1))
        ModelFactory(model_group_rel=mg3, train_end_time=datetime(2016, 1, 1))
        ModelFactory(model_group_rel=mg3, train_end_time=datetime(2017, 1, 1))
        # model group 4 only has three timestamps, should not pass
        ModelFactory(model_group_rel=mg4, train_end_time=datetime(2015, 1, 1))
        ModelFactory(model_group_rel=mg4, train_end_time=datetime(2016, 1, 1))

        session.commit()
        train_end_times = ['2014-01-01', '2015-01-01', '2016-01-01', '2017-01-01']
        model_groups = [1, 2, 3, 4]
        model_group_ids = model_groups_filter(
             train_end_times=train_end_times,
             initial_model_group_ids=model_groups,
             models_table='models',
             db_engine=engine
        )

        return model_group_ids

    def test_have_same_train_end_times(self):
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            pass_model_groups = self.filter_same_train_end_times(engine)
            assert pass_model_groups == {1, 3}


class ModelGroupThresholderTest(TestCase):

    metric_filters = [
        {
            'metric': 'precision@',
            'parameter': '100_abs',
            'max_from_best': 0.2,
            'threshold_value': 0.4,
        },
        {
            'metric': 'recall@',
            'parameter': '100_abs',
            'max_from_best': 0.2,
            'threshold_value': 0.4,
        },
        {
            'metric': 'false positives@',
            'parameter': '100_abs',
            'max_from_best': 30,
            'threshold_value': 50,
        }
    ]

    def setup_data(self, engine):
        ensure_db(engine)
        init_engine(engine)
        ModelGroupFactory(model_group_id=1, model_type='modelType1')
        ModelGroupFactory(model_group_id=2, model_type='modelType2')
        ModelGroupFactory(model_group_id=3, model_type='modelType3')
        ModelGroupFactory(model_group_id=4, model_type='modelType4')
        ModelGroupFactory(model_group_id=5, model_type='modelType5')
        session.commit()
        distance_table = DistanceFromBestTable(
            db_engine=engine,
            models_table='models',
            distance_table='dist_table'
        )
        distance_table._create()
        distance_rows = [
            # 2014: model group 1 should pass both close and min checks
            (1, 1, '2014-01-01', 'precision@', '100_abs', 0.5, 0.5, 0.0, 0.38),
            (1, 1, '2014-01-01', 'recall@', '100_abs', 0.5, 0.5, 0.0, 0.38),
            (1, 1, '2014-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # 2015: model group 1 should not pass close check
            (1, 2, '2015-01-01', 'precision@', '100_abs', 0.5, 0.88, 0.38, 0.0),
            (1, 2, '2015-01-01', 'recall@', '100_abs', 0.5, 0.88, 0.38, 0.0),
            (1, 2, '2015-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            (1, 3, '2016-01-01', 'precision@', '100_abs', 0.46, 0.46, 0.0, 0.11),
            (1, 3, '2016-01-01', 'recall@', '100_abs', 0.46, 0.46, 0.0, 0.11),
            (1, 3, '2016-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # 2014: model group 2 should not pass min check
            (2, 4, '2014-01-01', 'precision@', '100_abs', 0.39, 0.5, 0.11, 0.5),
            (2, 4, '2014-01-01', 'recall@', '100_abs', 0.5, 0.5, 0.0, 0.38),
            (2, 4, '2014-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # 2015: model group 2 should pass both checks
            (2, 5, '2015-01-01', 'precision@', '100_abs', 0.69, 0.88, 0.19, 0.12),
            (2, 5, '2015-01-01', 'recall@', '100_abs', 0.69, 0.88, 0.19, 0.0),
            (2, 5, '2015-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            (2, 6, '2016-01-01', 'precision@', '100_abs', 0.34, 0.46, 0.12, 0.11),
            (2, 6, '2016-01-01', 'recall@', '100_abs', 0.46, 0.46, 0.0, 0.11),
            (2, 6, '2016-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # model group 3 not included in this round
            (3, 7, '2014-01-01', 'precision@', '100_abs', 0.28, 0.5, 0.22, 0.0),
            (3, 7, '2014-01-01', 'recall@', '100_abs', 0.5, 0.5, 0.0, 0.38),
            (3, 7, '2014-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            (3, 8, '2015-01-01', 'precision@', '100_abs', 0.88, 0.88, 0.0, 0.02),
            (3, 8, '2015-01-01', 'recall@', '100_abs', 0.5, 0.88, 0.38, 0.0),
            (3, 8, '2015-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            (3, 9, '2016-01-01', 'precision@', '100_abs', 0.44, 0.46, 0.02, 0.11),
            (3, 9, '2016-01-01', 'recall@', '100_abs', 0.46, 0.46, 0.0, 0.11),
            (3, 9, '2016-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # 2014: model group 4 should not pass any checks
            (4, 10, '2014-01-01', 'precision@', '100_abs', 0.29, 0.5, 0.21, 0.21),
            (4, 10, '2014-01-01', 'recall@', '100_abs', 0.5, 0.5, 0.0, 0.38),
            (4, 10, '2014-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # 2015: model group 4 should not pass close check
            (4, 11, '2015-01-01', 'precision@', '100_abs', 0.67, 0.88, 0.21, 0.21),
            (4, 11, '2015-01-01', 'recall@', '100_abs', 0.5, 0.88, 0.38, 0.0),
            (4, 11, '2015-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            (4, 12, '2016-01-01', 'precision@', '100_abs', 0.25, 0.46, 0.21, 0.21),
            (4, 12, '2016-01-01', 'recall@', '100_abs', 0.46, 0.46, 0.0, 0.11),
            (4, 12, '2016-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # 2014: model group 5 should not pass because precision is good but not recall
            (5, 13, '2014-01-01', 'precision@', '100_abs', 0.5, 0.38, 0.0, 0.38),
            (5, 13, '2014-01-01', 'recall@', '100_abs', 0.3, 0.5, 0.2, 0.38),
            (5, 13, '2014-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # 2015: model group 5 should not pass because precision is good but not recall
            (5, 14, '2015-01-01', 'precision@', '100_abs', 0.5, 0.88, 0.38, 0.0),
            (5, 14, '2015-01-01', 'recall@', '100_abs', 0.3, 0.88, 0.58, 0.0),
            (5, 14, '2015-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            (5, 15, '2016-01-01', 'precision@', '100_abs', 0.46, 0.46, 0.0, 0.11),
            (5, 15, '2016-01-01', 'recall@', '100_abs', 0.3, 0.46, 0.16, 0.11),
            (5, 15, '2016-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
            # 2014: model group 6 is failed by false positives
            (6, 16, '2014-01-01', 'precision@', '100_abs', 0.5, 0.5, 0.0, 0.38),
            (6, 16, '2014-01-01', 'recall@', '100_abs', 0.5, 0.5, 0.0, 0.38),
            (6, 16, '2014-01-01', 'false positives@', '100_abs', 60, 30, 30, 10),
            # 2015: model group 6 is failed by false positives
            (6, 17, '2015-01-01', 'precision@', '100_abs', 0.5, 0.88, 0.38, 0.0),
            (6, 17, '2015-01-01', 'recall@', '100_abs', 0.5, 0.38, 0.0, 0.38),
            (6, 17, '2015-01-01', 'false positives@', '100_abs', 60, 30, 30, 10),
            (6, 18, '2016-01-01', 'precision@', '100_abs', 0.46, 0.46, 0.0, 0.11),
            (6, 18, '2016-01-01', 'recall@', '100_abs', 0.5, 0.5, 0.0, 0.38),
            (6, 18, '2016-01-01', 'false positives@', '100_abs', 40, 30, 10, 10),
        ]
        for dist_row in distance_rows:
            engine.execute(
                'insert into dist_table values (%s, %s, %s, %s, %s, %s, %s, %s, %s)',
                dist_row
            )
        thresholder = ModelGroupThresholder(
            distance_from_best_table=distance_table,
            train_end_times=['2014-01-01', '2015-01-01'],
            initial_model_group_ids=[1, 2, 4, 5, 6],
            initial_metric_filters=self.metric_filters
        )
        return thresholder

    def dataframe_as_of(self, thresholder, train_end_time):
        return thresholder.distance_from_best_table.dataframe_as_of(
            model_group_ids=thresholder._initial_model_group_ids,
            train_end_time=train_end_time,
        )

    def test_thresholder_2014_close(self):
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            thresholder = self.setup_data(engine)

            assert thresholder.model_groups_close_to_best_case(
                self.dataframe_as_of(thresholder, '2014-01-01')
            ) == set([1, 2])

    def test_thresholder_2015_close(self):
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            thresholder = self.setup_data(engine)
            assert thresholder.model_groups_close_to_best_case(
                self.dataframe_as_of(thresholder, '2015-01-01')
            ) == set([2])

    def test_thresholder_2014_threshold(self):
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            thresholder = self.setup_data(engine)
            assert thresholder.model_groups_past_threshold(
                self.dataframe_as_of(thresholder, '2014-01-01')
            ) == set([1])

    def test_thresholder_2015_threshold(self):
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            thresholder = self.setup_data(engine)
            assert thresholder.model_groups_past_threshold(
                self.dataframe_as_of(thresholder, '2015-01-01')
            ) == set([1, 2, 4])

    def test_thresholder_all_rules(self):
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            thresholder = self.setup_data(engine)
            # The multi-date version of this function should have
            # the mins ANDed together and the closes ORed together
            assert thresholder.model_groups_passing_rules() == set([1])

    def test_update_filters(self):
        with testing.postgresql.Postgresql() as postgresql:
            engine = create_engine(postgresql.url())
            thresholder = self.setup_data(engine)
            assert thresholder.model_group_ids == set([1])
            thresholder.update_filters([])
            assert thresholder.model_group_ids == set([1, 2, 4, 5, 6])



import yaml
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL
from sqlalchemy.pool import QueuePool

from triage.component.results_schema import Base


def ensure_db(engine):
    Base.metadata.create_all(engine)


def connect(poolclass=QueuePool):
    with open('database.yaml') as fd:
        config = yaml.load(fd)
        dburl = URL(
            'postgres',
            host=config['host'],
            username=config['user'],
            database=config['db'],
            password=config['pass'],
            port=config['port'],
        )
        return create_engine(dburl, poolclass=poolclass)



import logging
import os
from abc import ABC, abstractmethod

from descriptors import cachedproperty
from timeout import timeout
from sqlalchemy.engine import Engine

from triage.component.architect.label_generators import (
    LabelGenerator,
    LabelGeneratorNoOp,
    DEFAULT_LABEL_NAME,
)

from triage.component.architect.features import (
    FeatureGenerator,
    FeatureDictionaryCreator,
    FeatureGroupCreator,
    FeatureGroupMixer,
)
from triage.component.architect.planner import Planner
from triage.component.architect.builders import MatrixBuilder
from triage.component.architect.state_table_generators import (
    StateTableGeneratorFromDense,
    StateTableGeneratorFromEntities,
    StateTableGeneratorFromQuery,
    StateTableGeneratorNoOp
)
from triage.component.timechop import Timechop
from triage.component.results_schema import upgrade_db
from triage.component.catwalk.model_grouping import ModelGrouper
from triage.component.catwalk.model_trainers import ModelTrainer
from triage.component.catwalk.model_testers import ModelTester
from triage.component.catwalk.utils import save_experiment_and_get_hash
from triage.component.catwalk.storage import CSVMatrixStore, ModelStorageEngine, ProjectStorage, MatrixStorageEngine

from triage.experiments import CONFIG_VERSION
from triage.experiments.validate import ExperimentValidator

from triage.database_reflection import table_has_data
from triage.util.conf import dt_from_str
from triage.util.db import create_engine


class ExperimentBase(ABC):
    """The base class for all Experiments.

    Subclasses must implement the following four methods:
    process_query_tasks
    process_matrix_build_tasks
    process_train_tasks
    process_model_test_tasks

    Look at singlethreaded.py for reference implementation of each.

    Args:
        config (dict)
        db_engine (triage.util.db.SerializableDbEngine or sqlalchemy.engine.Engine)
        project_path (string)
        replace (bool)
        cleanup_timeout (int)
    """
    cleanup_timeout = 60  # seconds

    def __init__(
        self,
        config,
        db_engine,
        project_path=None,
        matrix_storage_class=CSVMatrixStore,
        replace=True,
        cleanup=False,
        cleanup_timeout=None,
    ):
        self._check_config_version(config)
        self.config = config

        if isinstance(db_engine, Engine):
            logging.warning('Raw, unserializable SQLAlchemy engine passed. URL will be used, other options may be lost in multi-process environments')
            self.db_engine = create_engine(db_engine.url)
        else:
            self.db_engine = db_engine

        self.project_storage = ProjectStorage(project_path)
        self.model_storage_engine = ModelStorageEngine(self.project_storage)
        self.matrix_storage_engine = MatrixStorageEngine(self.project_storage, matrix_storage_class)
        self.project_path = project_path
        self.replace = replace
        upgrade_db(db_engine=self.db_engine)

        self.features_schema_name = 'features'
        self.experiment_hash = save_experiment_and_get_hash(self.config,
                                                            self.db_engine)
        self.labels_table_name = 'labels_{}'.format(self.experiment_hash)
        self.initialize_components()

        self.cleanup = cleanup
        if self.cleanup:
            logging.info('cleanup is set to True, so intermediate tables (labels and states) will be removed after matrix creation')
        else:
            logging.info('cleanup is set to False, so intermediate tables (labels and states) will not be removed after matrix creation')
        self.cleanup_timeout = (self.cleanup_timeout if cleanup_timeout is None
                                else cleanup_timeout)

    def _check_config_version(self, config):
        if 'config_version' in config:
            config_version = config['config_version']
        else:
            logging.warning('config_version key not found in experiment config. '
                            'Assuming v1, which may not be correct')
            config_version = 'v1'
        if config_version != CONFIG_VERSION:
            raise ValueError(
                "Experiment config '{}' "
                "does not match current version '{}'. "
                "Will not run experiment."
                .format(config_version, CONFIG_VERSION)
            )

    def initialize_components(self):
        split_config = self.config['temporal_config']

        self.chopper = Timechop(**split_config)

        cohort_config = self.config.get('cohort_config', {})
        if 'query' in cohort_config:
            self.state_table_generator = StateTableGeneratorFromQuery(
                experiment_hash=self.experiment_hash,
                db_engine=self.db_engine,
                query=cohort_config['query']
            )
        elif 'entities_table' in cohort_config:
            self.state_table_generator = StateTableGeneratorFromEntities(
                experiment_hash=self.experiment_hash,
                db_engine=self.db_engine,
                entities_table=cohort_config['entities_table']
            )
        elif 'dense_states' in cohort_config:
            self.state_table_generator = StateTableGeneratorFromDense(
                experiment_hash=self.experiment_hash,
                db_engine=self.db_engine,
                dense_state_table=cohort_config['dense_states']['table_name']
            )
        else:
            logging.warning('cohort_config missing or unrecognized. Without a cohort, you will not be able to make matrices or perform feature imputation.')
            self.state_table_generator = StateTableGeneratorNoOp()

        if 'label_config' in self.config:
            self.label_generator = LabelGenerator(
                label_name=self.config['label_config'].get('name', None),
                query=self.config['label_config']['query'],
                db_engine=self.db_engine,
            )
        else:
            self.label_generator = LabelGeneratorNoOp()
            logging.warning('label_config missing or unrecognized. Without labels, you will not be able to make matrices.')

        self.feature_dictionary_creator = FeatureDictionaryCreator(
            features_schema_name=self.features_schema_name,
            db_engine=self.db_engine,
        )

        self.feature_generator = FeatureGenerator(
            features_schema_name=self.features_schema_name,
            replace=self.replace,
            db_engine=self.db_engine,
            feature_start_time=split_config['feature_start_time']
        )

        self.feature_group_creator = FeatureGroupCreator(
            self.config.get('feature_group_definition', {'all': [True]})
        )

        self.feature_group_mixer = FeatureGroupMixer(
            self.config.get('feature_group_strategies', ['all'])
        )

        self.planner = Planner(
            feature_start_time=dt_from_str(split_config['feature_start_time']),
            label_names=[self.config.get('label_config', {}).get('name', DEFAULT_LABEL_NAME)],
            label_types=['binary'],
            cohort_name=self.config.get('cohort_config', {}).get('name', None),
            states=self.config.get('cohort_config', {}).get('dense_states', {})
            .get('state_filters', []),
            user_metadata=self.config.get('user_metadata', {}),
        )

        self.matrix_builder = MatrixBuilder(
            db_config={
                'features_schema_name': self.features_schema_name,
                'labels_schema_name': 'public',
                'labels_table_name': self.labels_table_name,
                # TODO: have planner/builder take state table later on, so we
                # can grab it from the StateTableGenerator instead of
                # duplicating it here
                'sparse_state_table_name': self.sparse_states_table_name,
            },
            matrix_storage_engine=self.matrix_storage_engine,
            include_missing_labels_in_train_as=self.config.get('label_config', {})
            .get('include_missing_labels_in_train_as', None),
            engine=self.db_engine,
            replace=self.replace
        )

        self.trainer = ModelTrainer(
            experiment_hash=self.experiment_hash,
            model_storage_engine=self.model_storage_engine,
            model_grouper=ModelGrouper(self.config.get('model_group_keys', [])),
            db_engine=self.db_engine,
            replace=self.replace
        )

        self.tester = ModelTester(
            model_storage_engine=self.model_storage_engine,
            matrix_storage_engine=self.matrix_storage_engine,
            replace=self.replace,
            db_engine=self.db_engine,
            individual_importance_config=self.config.get('individual_importance', {}),
            evaluator_config=self.config.get('scoring', {})
        )

    @property
    def sparse_states_table_name(self):
        return 'tmp_sparse_states_{}'.format(self.experiment_hash)

    @cachedproperty
    def split_definitions(self):
        """Temporal splits based on the experiment's configuration

        Returns: (dict) temporal splits

        Example:
        ```
        {
            'feature_start_time': {datetime},
            'feature_end_time': {datetime},
            'label_start_time': {datetime},
            'label_end_time': {datetime},
            'train_matrix': {
                'first_as_of_time': {datetime},
                'last_as_of_time': {datetime},
                'matrix_info_end_time': {datetime},
                'training_label_timespan': {str},
                'training_as_of_date_frequency': {str},
                'max_training_history': {str},
                'as_of_times': [list of {datetime}s]
            },
            'test_matrices': [list of matrix defs similar to train_matrix]
        }
        ```

        (When updating/setting split definitions, matrices should have
        UUIDs.)

        """
        split_definitions = self.chopper.chop_time()
        logging.info('Computed and stored split definitions: %s',
                     split_definitions)
        logging.info('\n----TIME SPLIT SUMMARY----\n')
        logging.info('Number of time splits: {}'.format(len(split_definitions)))
        for split_index, split in enumerate(split_definitions):
            train_times = split['train_matrix']['as_of_times']
            test_times = [as_of_time for test_matrix in split['test_matrices']
                          for as_of_time in test_matrix['as_of_times']]
            logging.info('''Split index {}:
            Training as_of_time_range: {} to {} ({} total)
            Testing as_of_time range: {} to {} ({} total)\n\n'''.format(
                split_index,
                min(train_times),
                max(train_times),
                len(train_times),
                min(test_times),
                max(test_times),
                len(test_times)
            ))

        return split_definitions

    @cachedproperty
    def all_as_of_times(self):
        """All 'as of times' in experiment config

        Used for label and feature generation.

        Returns: (list) of datetimes

        """
        all_as_of_times = []
        for split in self.split_definitions:
            all_as_of_times.extend(split['train_matrix']['as_of_times'])
            logging.debug('Adding as_of_times from train matrix: %s',
                         split['train_matrix']['as_of_times'])
            for test_matrix in split['test_matrices']:
                logging.debug('Adding as_of_times from test matrix: %s',
                             test_matrix['as_of_times'])
                all_as_of_times.extend(test_matrix['as_of_times'])

        logging.info(
            'Computed %s total as_of_times for label and feature generation',
            len(all_as_of_times)
        )
        distinct_as_of_times = list(set(all_as_of_times))
        logging.info(
            'Computed %s distinct as_of_times for label and feature generation',
            len(distinct_as_of_times)
        )
        logging.info('You can view all as_of_times by inspecting `.all_as_of_times` on this Experiment')
        return distinct_as_of_times

    @cachedproperty
    def collate_aggregations(self):
        """Collation of ``Aggregation`` objects used by this experiment.

        Returns: (list) of ``collate.Aggregation`` objects

        """
        logging.info('Creating collate aggregations')
        cohort_table = self.state_table_generator.sparse_table_name
        if 'feature_aggregations' not in self.config:
            logging.warning('No feature_aggregation config is available')
            return []
        return self.feature_generator.aggregations(
            feature_aggregation_config=self.config['feature_aggregations'],
            feature_dates=self.all_as_of_times,
            state_table=cohort_table
        )

    @cachedproperty
    def feature_aggregation_table_tasks(self):
        """All feature table query tasks specified by this
        ``Experiment``.

        Returns: (dict) keys are group table names, values are
            themselves dicts, each with keys for different stages of
            table creation (prepare, inserts, finalize) and with values
            being lists of SQL commands

        """
        logging.info('Calculating feature tasks for %s as_of_times',
                     len(self.all_as_of_times))
        return self.feature_generator.generate_all_table_tasks(
            self.collate_aggregations,
            task_type='aggregation'
        )

    @cachedproperty
    def feature_imputation_table_tasks(self):
        """All feature imputation query tasks specified by this
        ``Experiment``.

        Returns: (dict) keys are group table names, values are
            themselves dicts, each with keys for different stages of
            table creation (prepare, inserts, finalize) and with values
            being lists of SQL commands

        """
        logging.info('Calculating feature tasks for %s as_of_times',
                     len(self.all_as_of_times))
        return self.feature_generator.generate_all_table_tasks(
            self.collate_aggregations,
            task_type='imputation'
        )

    @cachedproperty
    def master_feature_dictionary(self):
        """All possible features found in the database. Not all features
        will necessarily end up in matrices

        Returns: (list) of dicts, keys being feature table names and
        values being lists of feature names

        """
        result = self.feature_dictionary_creator.feature_dictionary(
            feature_table_names=self.feature_imputation_table_tasks.keys(),
            index_column_lookup=self.feature_generator.index_column_lookup(
                self.collate_aggregations
            )
        )
        logging.info('Computed master feature dictionary: %s', result)
        return result

    @property
    def feature_dicts(self):
        """Feature dictionaries, representing the feature tables and
        columns configured in this experiment after computing feature
        groups.

        Returns: (list) of dicts, keys being feature table names and
        values being lists of feature names

        """
        return self.feature_group_mixer.generate(
            self.feature_group_creator.subsets(self.master_feature_dictionary)
        )

    @cachedproperty
    def matrix_build_tasks(self):
        """Tasks for all matrices that need to be built as a part of
        this Experiment.

        Each task contains arguments understood by
        ``Architect.build_matrix``.

        Returns: (list) of dicts

        """
        if not table_has_data(self.sparse_states_table_name, self.db_engine):
            logging.warning('cohort table is not populated, cannot build any matrices')
            return {}
        if not table_has_data(self.labels_table_name, self.db_engine):
            logging.warning('labels table is not populated, cannot build any matrices')
            return {}
        (
            updated_split_definitions,
            matrix_build_tasks
        ) = self.planner.generate_plans(
            self.split_definitions,
            self.feature_dicts
        )
        self.full_matrix_definitions = updated_split_definitions
        return matrix_build_tasks

    @cachedproperty
    def full_matrix_definitions(self):
        """Full matrix definitions

        Returns: (list) temporal and feature information for each matrix

        """
        (
            updated_split_definitions,
            matrix_build_tasks
        ) = self.planner.generate_plans(
            self.split_definitions,
            self.feature_dicts
        )
        self.matrix_build_tasks = matrix_build_tasks
        return updated_split_definitions

    @property
    def all_label_timespans(self):
        """All train and test label timespans

        Returns: (list) label timespans, in string form as they appeared in the experiment config

        """
        return list(set(
            self.config['temporal_config']['training_label_timespans'] +
            self.config['temporal_config']['test_label_timespans']
        ))

    def generate_labels(self):
        """Generate labels based on experiment configuration

        Results are stored in the database, not returned
        """
        self.label_generator.generate_all_labels(
            self.labels_table_name,
            self.all_as_of_times,
            self.all_label_timespans
        )

    def generate_cohort(self):
        self.state_table_generator.generate_sparse_table(
            as_of_dates=self.all_as_of_times
        )

    def log_split(self, split_num, split):
        logging.info(
            'Starting train/test for %s out of %s: train range: %s to %s',
            split_num+1,
            len(self.full_matrix_definitions),
            split['train_matrix']['first_as_of_time'],
            split['train_matrix']['matrix_info_end_time'],
        )

    @abstractmethod
    def process_train_tasks(self, train_tasks):
        pass

    @abstractmethod
    def process_query_tasks(self, query_tasks):
        pass

    @abstractmethod
    def process_matrix_build_tasks(self, matrix_build_tasks):
        pass

    def generate_preimputation_features(self):
        self.process_query_tasks(self.feature_aggregation_table_tasks)
        logging.info('Finished running preimputation feature queries. The final results are in tables: %s',
                     ','.join(agg.get_table_name() for agg in self.collate_aggregations)
                     )

    def impute_missing_features(self):
        self.process_query_tasks(self.feature_imputation_table_tasks)
        logging.info('Finished running postimputation feature queries. The final results are in tables: %s',
                     ','.join(agg.get_table_name(imputed=True) for agg in self.collate_aggregations)
                     )

    def build_matrices(self):
        self.process_matrix_build_tasks(self.matrix_build_tasks)

    def generate_matrices(self):
        logging.info('Creating cohort')
        self.generate_cohort()
        logging.info('Creating labels')
        self.generate_labels()
        logging.info('Creating feature aggregation tables')
        self.generate_preimputation_features()
        logging.info('Creating feature imputation tables')
        self.impute_missing_features()
        logging.info('Building all matrices')
        self.build_matrices()

    def train_and_test_models(self):
        if 'grid_config' not in self.config:
            logging.warning('No grid_config was passed in the experiment config. No models will be trained')
            return

        for split_num, split in enumerate(self.full_matrix_definitions):
            self.log_split(split_num, split)
            train_store = self.matrix_storage_engine.get_store(split['train_uuid'])
            if train_store.empty:
                logging.warning('''Train matrix for split %s was empty,
                no point in training this model. Skipping
                ''', split['train_uuid'])
                continue
            if len(train_store.labels().unique()) == 1:
                logging.warning('''Train Matrix for split %s had only one
                unique value, no point in training this model. Skipping
                ''', split['train_uuid'])
                continue

            logging.info('Training models')

            train_tasks = self.trainer.generate_train_tasks(
                grid_config=self.config['grid_config'],
                misc_db_parameters=dict(
                    test=False,
                    model_comment=self.config.get('model_comment', None),
                ),
                matrix_store=train_store
            )
            model_ids = self.process_train_tasks(train_tasks)

            logging.info('Done training models for split %s', split_num)

            test_tasks = self.tester.generate_model_test_tasks(
                split=split,
                train_store=train_store,
                model_ids=model_ids,
            )
            logging.info('Found %s non-empty test matrices for split %s', len(test_tasks), split_num)

            self.process_model_test_tasks(test_tasks)

    def validate(self, strict=True):
        ExperimentValidator(self.db_engine, strict=strict).run(self.config)

    def _run(self):
        try:
            logging.info('Generating matrices')
            self.generate_matrices()
        finally:
            if self.cleanup:
                self.clean_up_tables()

        self.train_and_test_models()

    def clean_up_tables(self):
        logging.info('Cleaning up state and labels tables')
        with timeout(self.cleanup_timeout):
            self.state_table_generator.clean_up()
            self.label_generator.clean_up(self.labels_table_name)

    def run(self):
        try:
            self._run()
        except Exception:
            logging.exception('Run interrupted by uncaught exception')
            raise

    __call__ = run
